from __future__ import annotations

from dissect.volume.disk.schemes.apm import APM
from dissect.volume.disk.schemes.bsd import BSD
from dissect.volume.disk.schemes.gpt import GPT
from dissect.volume.disk.schemes.mbr import MBR

__all__ = [
    "APM",
    "BSD",
    "GPT",
    "MBR",
]



from __future__ import annotations

from dissect.cstruct import cstruct

md_def = """
#define MD_RESERVED_BYTES               (64 * 1024)
#define MD_RESERVED_SECTORS             (MD_RESERVED_BYTES / 512)

/*
 * RAID superblock.
 *
 * The RAID superblock maintains some statistics on each RAID configuration.
 * Each real device in the RAID set contains it near the end of the device.
 * Some of the ideas are copied from the ext2fs implementation.
 *
 * We currently use 4096 bytes as follows:
 *
 *  word offset     function
 *
 *     0  -    31   Constant generic RAID device information.
 *    32  -    63   Generic state information.
 *    64  -   127   Personality specific information.
 *   128  -   511   12 32-words descriptors of the disks in the raid set.
 *   512  -   911   Reserved.
 *   912  -  1023   Disk specific descriptor.
 */

#define MD_SB_GENERIC_CONSTANT_WORDS    32
#define MD_SB_GENERIC_STATE_WORDS       32
#define MD_SB_GENERIC_WORDS             (MD_SB_GENERIC_CONSTANT_WORDS + MD_SB_GENERIC_STATE_WORDS)
#define MD_SB_PERSONALITY_WORDS         64
#define MD_SB_DESCRIPTOR_WORDS          32
#define MD_SB_DISKS                     27
#define MD_SB_DISKS_WORDS               (MD_SB_DISKS * MD_SB_DESCRIPTOR_WORDS)
#define MD_SB_RESERVED_WORDS            (1024 - MD_SB_GENERIC_WORDS - MD_SB_PERSONALITY_WORDS - MD_SB_DISKS_WORDS - MD_SB_DESCRIPTOR_WORDS)

/*
 * Device "operational" state bits
 */
#define MD_DISK_FAULTY                  0           /* disk is faulty / operational */
#define MD_DISK_ACTIVE                  1           /* disk is running or spare disk */
#define MD_DISK_SYNC                    2           /* disk is in sync with the raid set */
#define MD_DISK_REMOVED                 3           /* disk is in sync with the raid set */
#define MD_DISK_CLUSTER_ADD             4           /* Initiate a disk add across the cluster
                                                     * For clustered enviroments only.
                                                     */
#define MD_DISK_CANDIDATE               5           /* disk is added as spare (local) until confirmed
                                                     * For clustered enviroments only.
                                                     */
#define MD_DISK_FAILFAST                10          /* Send REQ_FAILFAST if there are multiple
                                                     * devices available - and don't try to
                                                     * correct read errors.
                                                     */

#define MD_DISK_WRITEMOSTLY             9           /* disk is "write-mostly" is RAID1 config.
                                                     * read requests will only be sent here in
                                                     * dire need
                                                     */
#define MD_DISK_JOURNAL                 18          /* disk is used as the write journal in RAID-5/6 */

#define MD_DISK_ROLE_SPARE              0xffff
#define MD_DISK_ROLE_FAULTY             0xfffe
#define MD_DISK_ROLE_JOURNAL            0xfffd
#define MD_DISK_ROLE_MAX                0xff00      /* max value of regular disk role */

typedef struct mdp_device_descriptor_s {
    uint32  number;                                 /* 0 Device number in the entire set */
    uint32  major;                                  /* 1 Device major number */
    uint32  minor;                                  /* 2 Device minor number */
    uint32  raid_disk;                              /* 3 The role of the device in the raid set */
    uint32  state;                                  /* 4 Operational state */
    uint32  reserved[MD_SB_DESCRIPTOR_WORDS - 5];
} mdp_disk_t;

#define MD_SB_MAGIC                     0xa92b4efc

typedef struct mdp_superblock_s {
    /*
     * Constant generic information
     */
    uint32  md_magic;                               /*  0 MD identifier */
    uint32  major_version;                          /*  1 major version to which the set conforms */
    uint32  minor_version;                          /*  2 minor version ... */
    uint32  patch_version;                          /*  3 patchlevel version ... */
    uint32  gvalid_words;                           /*  4 Number of used words in this section */
    char    set_uuid0[4];                           /*  5 Raid set identifier */
    uint32  ctime;                                  /*  6 Creation time */
    uint32  level;                                  /*  7 Raid personality */
    uint32  size;                                   /*  8 Apparent size of each individual disk */
    uint32  nr_disks;                               /*  9 total disks in the raid set */
    uint32  raid_disks;                             /* 10 disks in a fully functional raid set */
    uint32  md_minor;                               /* 11 preferred MD minor device number */
    uint32  not_persistent;                         /* 12 does it have a persistent superblock */
    char    set_uuid1[4];                           /* 13 Raid set identifier #2 */
    char    set_uuid2[4];                           /* 14 Raid set identifier #3 */
    char    set_uuid3[4];                           /* 15 Raid set identifier #4 */
    uint32  gstate_creserved[MD_SB_GENERIC_CONSTANT_WORDS - 16];

    /*
     * Generic state information
     */
    uint32  utime;                                  /*  0 Superblock update time */
    uint32  state;                                  /*  1 State bits (clean, ...) */
    uint32  active_disks;                           /*  2 Number of currently active disks */
    uint32  working_disks;                          /*  3 Number of working disks */
    uint32  failed_disks;                           /*  4 Number of failed disks */
    uint32  spare_disks;                            /*  5 Number of spare disks */
    uint32  sb_csum;                                /*  6 checksum of the whole superblock */
    uint32  events_lo;                              /*  7 low-order of superblock update count */
    uint32  events_hi;                              /*  8 high-order of superblock update count */
    uint32  cp_events_lo;                           /*  9 low-order of checkpoint update count */
    uint32  cp_events_hi;                           /* 10 high-order of checkpoint update count */
    uint32  recovery_cp;                            /* 11 recovery checkpoint sector count */
    /* There are only valid for minor_version > 90 */
    uint64  reshape_position;                       /* 12,13 next address in array-space for reshape */
    uint32  new_level;                              /* 14 new level we are reshaping to */
    uint32  delta_disks;                            /* 15 change in number of raid_disks */
    uint32  new_layout;                             /* 16 new layout */
    uint32  new_chunk;                              /* 17 new chunk size (bytes) */
    uint32  gstate_sreserved[MD_SB_GENERIC_STATE_WORDS - 18];

    /*
     * Personality information
     */
    uint32  layout;                                 /*  0 the array's physical layout */
    uint32  chunk_size;                             /*  1 chunk size in bytes */
    uint32  root_pv;                                /*  2 LV root PV */
    uint32  root_block;                             /*  3 LV root block */
    uint32  pstate_reserved[MD_SB_PERSONALITY_WORDS - 4];

    /*
     * Disks information
     */
    mdp_disk_t  disks[MD_SB_DISKS];

    /*
     * Reserved
     */
    uint32  reserved[MD_SB_RESERVED_WORDS];

    /*
     * Active descriptor
     */
    mdp_disk_t  this_disk;

} mdp_super_t;

#define WriteMostly1                    1           /* mask for writemostly flag in above */
#define FailFast1                       2           /* Should avoid retries and fixups and just fail */

/*
 * The version-1 superblock :
 * All numeric fields are little-endian.
 *
 * total size: 256 bytes plus 2 per device.
 *  1K allows 384 devices.
 */
struct mdp_superblock_1 {
    /* constant array information - 128 bytes */
    uint32  magic;                                  /* MD_SB_MAGIC: 0xa92b4efc - little endian */
    uint32  major_version;                          /* 1 */
    uint32  feature_map;                            /* bit 0 set if 'bitmap_offset' is meaningful */
    uint32  pad0;                                   /* always set to 0 when writing */

    char    set_uuid[16];                           /* user-space generated. */
    char    set_name[32];                           /* set and interpreted by user-space */

    uint64  ctime;                                  /* lo 40 bits are seconds, top 24 are microseconds or 0*/
    int32   level;                                  /* -4 (multipath), -1 (linear), 0,1,4,5 */
    uint32  layout;                                 /* only for raid5 and raid10 currently */
    uint64  size;                                   /* used size of component devices, in 512byte sectors */

    uint32  chunksize;                              /* in 512byte sectors */
    uint32  raid_disks;
    union {
        uint32  bitmap_offset;                      /* sectors after start of superblock that bitmap starts
                                                     * NOTE: signed, so bitmap can be before superblock
                                                     * only meaningful of feature_map[0] is set.
                                                     */

        /* only meaningful when feature_map[MD_FEATURE_PPL] is set */
        struct {
            uint16  offset;                         /* sectors from start of superblock that ppl starts (signed) */
            uint16  size;                           /* ppl size in sectors */
        } ppl;
    };

    /* These are only valid with feature bit '4' */
    uint32  new_level;                              /* new level we are reshaping to */
    uint64  reshape_position;                       /* next address in array-space for reshape */
    uint32  delta_disks;                            /* change in number of raid_disks */
    uint32  new_layout;                             /* new layout */
    uint32  new_chunk;                              /* new chunk size (512byte sectors) */
    uint32  new_offset;                             /* signed number to add to data_offset in new
                                                     * layout. 0 == no-change. This can be
                                                     * different on each device in the array.
                                                     */

    /* constant this-device information - 64 bytes */
    uint64  data_offset;                            /* sector start of data, often 0 */
    uint64  data_size;                              /* sectors in this device that can be used for data */
    uint64  super_offset;                           /* sector start of this superblock */
    union {
        uint64  recovery_offset;                    /* sectors before this offset (from data_offset) have been recovered */
        uint64  journal_tail;                       /* journal tail of journal device (from data_offset) */
    };
    uint32  dev_number;                             /* permanent identifier of this device - not role in raid */
    uint32  cnt_corrected_read;                     /* number of read errors that were corrected by re-writing */
    char    device_uuid[16];                        /* user-space setable, ignored by kernel */
    uint8   devflags;                               /* per-device flags. Only two defined... */

    /* Bad block log.  If there are any bad blocks the feature flag is set.
     * If offset and size are non-zero, that space is reserved and available
     */
    uint8   bblog_shift;                            /* shift from sectors to block size */
    uint16  bblog_size;                             /* number of sectors reserved for list */
    int32   bblog_offset;                           /* sector offset from superblock to bblog, signed - not unsigned */

    /* array state information - 64 bytes */
    uint64  utime;                                  /* 40 bits second, 24 bits microseconds */
    uint64  events;                                 /* incremented when superblock updated */
    uint64  resync_offset;                          /* data before this offset (from data_offset) known to be in sync */
    uint32  sb_csum;                                /* checksum up to devs[max_dev] */
    uint32  max_dev;                                /* size of devs[] array to consider */
    char    pad3[64-32];                            /* set to 0 when writing */

    /* device state information. Indexed by dev_number.
     * 2 bytes per device
     * Note there are no per-device state flags. State information is rolled
     * into the 'roles' value.  If a device is spare or faulty, then it doesn't
     * have a meaningful role.
     */
    uint16  dev_roles[max_dev];                     /* role in array, or 0xffff for a spare, or 0xfffe for faulty */
};

/* non-obvious values for 'level' */
#define LEVEL_MULTIPATH                 (-4)
#define LEVEL_LINEAR                    (-1)
#define LEVEL_FAULTY                    (-5)

/*
 * Our supported algorithms
 */
#define ALGORITHM_LEFT_ASYMMETRIC       0           /* Rotating Parity N with Data Restart */
#define ALGORITHM_RIGHT_ASYMMETRIC      1           /* Rotating Parity 0 with Data Restart */
#define ALGORITHM_LEFT_SYMMETRIC        2           /* Rotating Parity N with Data Continuation */
#define ALGORITHM_RIGHT_SYMMETRIC       3           /* Rotating Parity 0 with Data Continuation */

/* Define non-rotating (raid4) algorithms.  These allow
 * conversion of raid4 to raid5.
 */
#define ALGORITHM_PARITY_0              4           /* P or P,Q are initial devices */
#define ALGORITHM_PARITY_N              5           /* P or P,Q are final devices. */

/* DDF RAID6 layouts differ from md/raid6 layouts in two ways.
 * Firstly, the exact positioning of the parity block is slightly
 * different between the 'LEFT_*' modes of md and the "_N_*" modes
 * of DDF.
 * Secondly, or order of datablocks over which the Q syndrome is computed
 * is different.
 * Consequently we have different layouts for DDF/raid6 than md/raid6.
 * These layouts are from the DDFv1.2 spec.
 * Interestingly DDFv1.2-Errata-A does not specify N_CONTINUE but
 * leaves RLQ=3 as 'Vendor Specific'
 */

#define ALGORITHM_ROTATING_ZERO_RESTART 8           /* DDF PRL=6 RLQ=1 */
#define ALGORITHM_ROTATING_N_RESTART    9           /* DDF PRL=6 RLQ=2 */
#define ALGORITHM_ROTATING_N_CONTINUE   10          /* DDF PRL=6 RLQ=3 */

/* For every RAID5 algorithm we define a RAID6 algorithm
 * with exactly the same layout for data and parity, and
 * with the Q block always on the last device (N-1).
 * This allows trivial conversion from RAID5 to RAID6
 */
#define ALGORITHM_LEFT_ASYMMETRIC_6     16
#define ALGORITHM_RIGHT_ASYMMETRIC_6    17
#define ALGORITHM_LEFT_SYMMETRIC_6      18
#define ALGORITHM_RIGHT_SYMMETRIC_6     19
#define ALGORITHM_PARITY_0_6            20
#define ALGORITHM_PARITY_N_6            ALGORITHM_PARITY_N
"""  # noqa: E501

c_md = cstruct().load(md_def)

SECTOR_SIZE = 512



from __future__ import annotations

from dissect.cstruct import cstruct

# Structures are copied from:
# https://github.com/freebsd/freebsd-src/blob/f21a6a6a8fc59393173d9a537ed8cebbdbd6343c/sys/geom/vinum/geom_vinum_var.h

vinum_def = """
struct timeval {
    uint64 sec;
    uint64 usec;
};

typedef uint64 off_t;

/*
 * Slice header
 *
 * Vinum drives start with this structure:
 *
 *                                             Sector
 * |--------------------------------------|
 * |   PDP-11 memorial boot block         |      0
 * |--------------------------------------|
 * |   Disk label, maybe                  |      1
 * |--------------------------------------|
 * |   Slice definition  (vinum_hdr)      |      8
 * |--------------------------------------|
 * |                                      |
 * |   Configuration info, first copy     |      9
 * |                                      |
 * |--------------------------------------|
 * |                                      |
 * |   Configuration info, second copy    |      9 + size of config
 * |                                      |
 * |--------------------------------------|
 *

/* Sizes and offsets of our information. */
#define GV_HDR_OFFSET       4096    /* Offset of vinum header. */
#define GV_HDR_LEN          512     /* Size of vinum header. */
#define GV_CFG_OFFSET       4608    /* Offset of first config copy. */
#define GV_CFG_LEN          65536   /* Size of config copy. */

/* This is where the actual data starts. */
#define GV_DATA_START       (GV_CFG_LEN * 2 + GV_CFG_OFFSET)
/* #define GV_DATA_START       (GV_CFG_LEN * 2 + GV_HDR_LEN) */

#define GV_MAXDRIVENAME     32      /* Maximum length of a device name. */

/*
 * hostname is 256 bytes long, but we don't need to shlep multiple copies in
 * vinum.  We use the host name just to identify this system, and 32 bytes
 * should be ample for that purpose.
 */

#define GV_HOSTNAME_LEN     32
struct gv_label {
    char            sysname[GV_HOSTNAME_LEN];   /* System name at creation time. */
    char            name[GV_MAXDRIVENAME];      /* Our name of the drive. */
    struct timeval  date_of_birth;              /* The time it was created ... */
    struct timeval  last_update;                /* ... and the time of last update. */
    off_t           drive_size;                 /* Total size incl. headers. */
};

#define GV_OLD_MAGIC        0x494E2056494E4F00LL
#define GV_OLD_NOMAGIC      0x4E4F2056494E4F00LL
#define GV_MAGIC            0x56494E554D2D3100LL
#define GV_NOMAGIC          0x56494E554D2D2D00LL

/* The 'header' of each valid vinum drive. */
struct gv_hdr {
    uint64_t        magic;
    uint64_t        config_length;
    struct gv_label label;
} header;
"""

c_vinum = cstruct(endian=">").load(vinum_def)

# Not really needed as this size is hardcoded in the various GV_*_OFFSET and related values
SECTOR_SIZE = 512

MAGIC_ACTIVE = {c_vinum.GV_OLD_MAGIC, c_vinum.GV_MAGIC}
MAGIC_INACTIVE = {c_vinum.GV_OLD_NOMAGIC, c_vinum.GV_NOMAGIC}



from __future__ import annotations

from dissect.cstruct import cstruct

lvm_def = """
struct label_header {
    char    id[8];
    uint64  sector;
    uint32  crc;
    uint32  offset;
    char    type[8];
};

struct pv_header {
    char    pv_uuid[32];
    uint64  device_size;
};

struct disk_locn {
    uint64  offset;         /* Offset in bytes to start sector */
    uint64  size;           /* Bytes */
};

// Metadata area header
struct mda_header {
    uint32  checksum;       /* Checksum of rest of mda_header */
    char    magic[16];      /* To aid scans for metadata */
    uint32  version;
    uint64  start;          /* Absolute start byte of mda_header */
    uint64  size;           /* Size of metadata area */
};

struct raw_locn {
    uint64  offset;         /* Offset in bytes to start sector */
    uint64  size;           /* Bytes */
    uint32  checksum;
    uint32  flags;
};

#define RAW_LOCN_IGNORED    0x00000001
"""

c_lvm = cstruct().load(lvm_def)

SECTOR_SIZE = 512

LABEL_SCAN_SECTORS = 4

STATUS_FLAG_ALLOCATABLE = "ALLOCATABLE"  # pv only
STATUS_FLAG_RESIZEABLE = "RESIZEABLE"  # vg only
STATUS_FLAG_READ = "READ"
STATUS_FLAG_VISIBLE = "VISIBLE"  # lv only
STATUS_FLAG_WRITE = "WRITE"



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.volume.disk.schemes import APM, BSD, GPT, MBR
from dissect.volume.exceptions import DiskError

if TYPE_CHECKING:
    from dissect.volume.disk.partition import Partition


class Disk:
    def __init__(self, fh: BinaryIO, sector_size: int = 512):
        self.fh = fh
        self.sector_size = sector_size
        self.scheme: APM | GPT | MBR | BSD = None
        self.partitions: list[Partition] = []

        start = fh.tell()
        errors = []

        # The GPT scheme also parses the protective MBR, so it must be tried before MBR.
        # BSD is usually contained in another scheme's partition, but it can also live standalone.
        # We try to detect BSD as part of another scheme later on, so only try to detect BSD last
        # as standalone.
        for scheme in [GPT, MBR, APM, BSD]:
            try:
                fh.seek(start)
                self.scheme = scheme(fh, sector_size=self.sector_size)

                break
            except Exception as e:
                errors.append(str(e))

        if not self.scheme:
            raise DiskError("Unable to detect a valid partition scheme:\n- {}".format("\n- ".join(errors)))

        main_scheme = self.scheme
        for partition in main_scheme.partitions:
            # BSD disklabel can also be relative from any other partition
            if partition.type in BSD.TYPES:
                self.scheme = BSD(partition.open(), sector_size=self.sector_size)
                self.partitions.extend(self.scheme.partitions)
            else:
                self.partitions.append(partition)

        if isinstance(self.scheme, MBR) and any(p.type == 0xEE for p in self.partitions):
            raise DiskError("Found GPT type partition, but MBR scheme detected. Maybe 4K sector size.")

    @property
    def serial(self) -> int | None:
        if isinstance(self.scheme, MBR):
            return self.scheme.mbr.vol_no
        return None



from __future__ import annotations

from typing import BinaryIO

from dissect.util.stream import RangeStream

from dissect.volume.raid.stream import (
    Level,
    LinearStream,
    RAID0Stream,
    RAID10Stream,
    RAID456Stream,
)

DiskMap = dict[int, tuple[int, "PhysicalDisk"]]
"""A dict of the structure: {disk_idx: {data_offset: PhysicalDisk}}"""


class RAID:
    def __init__(self, configurations: list[Configuration]):
        self.configurations = configurations


class Configuration:
    def __init__(self, physical_disks: list[PhysicalDisk], virtual_disks: list[VirtualDisk]):
        self.physical_disks = physical_disks
        self.virtual_disks = virtual_disks


class PhysicalDisk:
    def __init__(self, fh: BinaryIO, offset: int, size: int):
        self.fh = fh
        self.offset = offset
        self.size = size

    def open(self) -> BinaryIO:
        """Return a file-like object of the data section of the disk."""
        return RangeStream(self.fh, self.offset, self.size)


class VirtualDisk:
    def __init__(
        self,
        name: str,
        uuid: str,
        size: int,
        level: int,
        layout: int,
        stripe_size: int,
        num_disks: int,
        disk_map: DiskMap,
    ):
        self.name = name
        self.uuid = uuid
        self.size = size
        self.level = level
        self.layout = layout
        self.stripe_size = stripe_size
        self.num_disks = num_disks
        self.disk_map = disk_map

    def open(self) -> BinaryIO:
        """Return a file-like object of the RAID volume in this set."""
        if self.level == Level.LINEAR:
            return LinearStream(self)
        if self.level == Level.RAID0:
            return RAID0Stream(self)
        if self.level == Level.RAID1:
            # Don't really care which mirror to read from, so just open the first disk we have.
            # We don't know if we have the first disk or the n-th disk so we select the first
            # available disk using an iterable.
            _, (_, disk) = next(iter(self.disk_map.items()))
            return disk.open()
        if self.level in (Level.RAID4, Level.RAID5, Level.RAID6):
            return RAID456Stream(self)
        if self.level == Level.RAID10:
            return RAID10Stream(self)
        raise NotImplementedError(f"Unsupported RAID level: {self.level}")



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts

from dissect.volume.ddf.c_ddf import c_ddf
from dissect.volume.exceptions import DDFError
from dissect.volume.raid.raid import (
    RAID,
    Configuration,
    PhysicalDisk,
    VirtualDisk,
)
from dissect.volume.raid.stream import Layout, Level

if TYPE_CHECKING:
    from dissect.volume.raid.raid import (
        DiskMap,
    )

DEFAULT_SECTOR_SIZE = 512

DECADE = 3600 * 24 * (365 * 10 + 2)


class DDF(RAID):
    def __init__(
        self,
        fh: list[BinaryIO | DDFPhysicalDisk] | BinaryIO | DDFPhysicalDisk,
        sector_size: int = DEFAULT_SECTOR_SIZE,
    ):
        fhs = [fh] if not isinstance(fh, list) else fh
        physical_disks = [DDFPhysicalDisk(f, sector_size) if not isinstance(f, DDFPhysicalDisk) else f for f in fhs]

        config_map = {}
        for pd in physical_disks:
            config_map.setdefault(pd.anchor.DDF_Header_GUID, []).append(pd)

        super().__init__([DDFConfiguration(disks) for disks in config_map.values()])


class DDFConfiguration(Configuration):
    def __init__(self, physical_disks: list[DDFPhysicalDisk]):
        pd_map: dict[int, DDFPhysicalDisk] = {}
        vde_map: dict[bytes, VirtualDiskRecord] = {}
        vdcr_map: dict[bytes, VirtualDiskConfigurationRecord] = {}
        vdcr_uniq: dict[tuple[bytes, int], VirtualDiskConfigurationRecord] = {}

        for pd in physical_disks:
            pd_map[pd.reference] = pd
            vde_map.update({vde.guid: vde for vde in pd.virtual_disk_records})
            vdcr_map.update({vdcr.guid: vdcr for vdcr in pd.virtual_disk_configuration_records})
            vdcr_uniq.update(
                {(vdcr.guid, vdcr.secondary_element_seq): vdcr for vdcr in pd.virtual_disk_configuration_records}
            )

        vd_map: dict = {}
        for vdcr in vdcr_uniq.values():
            vd_pd_map = vd_map.setdefault(vdcr.guid, {})

            count = vdcr.primary_element_count
            sec = 0 if vdcr.secondary_element_count == 1 else vdcr.secondary_element_seq

            i = 0
            for starting_block, pds in zip(vdcr.starting_block, vdcr.physical_disk_sequence, strict=False):
                if matched_pd := pd_map.get(pds):
                    vd_pd_map[sec * count + i] = (starting_block, matched_pd)

                if pds != 0xFFFFFFFF:
                    i += 1

        virtual_disks = [DDFVirtualDisk(vdcr_map[guid], vde_map[guid], vd_map[guid]) for guid in vd_map]
        super().__init__(physical_disks, virtual_disks)


class DDFVirtualDisk(VirtualDisk):
    def __init__(
        self,
        vdcr: VirtualDiskConfigurationRecord,
        vdr: VirtualDiskRecord,
        disks: DiskMap,
    ):
        self.vdcr = vdcr
        self.vdr = vdr
        self.disks = disks

        if (block_size := self.vdcr.block_size) == 0xFFFF:
            block_size = next(iter(disks.values()))[1].block_size

        level, layout, num_disks = _convert_raid_layout(
            vdcr.primary_raid_level,
            vdcr.raid_level_qualifier,
            vdcr.secondary_raid_level,
            vdcr.primary_element_count,
            vdcr.secondary_element_count,
        )

        super().__init__(
            self.vdr.name,
            self.vdr.guid.hex(),
            self.vdcr.size * block_size,
            level,
            layout,
            (2**self.vdcr.strip_size) * block_size,
            num_disks,
            disks,
        )

    @property
    def virtual_disk_configuration_record(self) -> VirtualDiskConfigurationRecord:
        return self.vdcr

    @property
    def virtual_disk_record(self) -> VirtualDiskRecord:
        return self.vdr


class DDFPhysicalDisk(PhysicalDisk):
    def __init__(self, fh: BinaryIO, sector_size: int = DEFAULT_SECTOR_SIZE):
        self.sector_size = sector_size

        fh.seek(-sector_size, io.SEEK_END)
        self.anchor = c_ddf.DDF_Header(fh)
        if self.anchor.Signature != c_ddf.DDF_HEADER_SIGNATURE:
            raise DDFError(
                "Invalid DDF anchor header. "
                f"Signature: {self.anchor.Signature:#010x}, expected {c_ddf.DDF_HEADER_SIGNATURE:#x}"
            )

        self.primary_header_offset = self.anchor.Primary_Header_LBA * sector_size
        self.secondary_header_offset = None

        fh.seek(self.primary_header_offset)
        self.primary_header = c_ddf.DDF_Header(fh)
        self.secondary_header = None
        if self.anchor.Secondary_Header_LBA != 0xFFFFFFFFFFFFFFFF:
            self.secondary_header_offset = self.anchor.Secondary_Header_LBA * sector_size
            fh.seek(self.secondary_header_offset)
            self.secondary_header = c_ddf.DDF_Header(fh)

        self.active_header_offset = self.primary_header_offset

        header_offset = self.active_header_offset
        fh.seek(header_offset + (self.anchor.Controller_Data_Section * sector_size))
        self.controller_data = ControllerData(fh)

        physical_disk_records_offset = header_offset + (self.anchor.Physical_Disk_Records_Section * sector_size)
        self.physical_disk_records = _read_physical_disk_records(fh, physical_disk_records_offset)

        virtual_disk_records_offset = header_offset + (self.anchor.Virtual_Disk_Records_Section * sector_size)
        self.virtual_disk_records = _read_virtual_disk_records(fh, virtual_disk_records_offset)

        configuration_records_offset = header_offset + (self.anchor.Configuration_Records_Section * sector_size)
        self.virtual_disk_configuration_records = _read_virtual_disk_configuration_records(
            fh,
            configuration_records_offset,
            self.anchor.Configuration_Records_Section_Length // self.anchor.Configuration_Record_Length,
            self.anchor.Configuration_Record_Length * sector_size,
            self.anchor.Max_Primary_Element_Entries,
        )

        fh.seek(header_offset + (self.anchor.Physical_Disk_Data_Section * sector_size))
        self.physical_disk_data = PhysicalDiskData(fh)

        self.guid = self.physical_disk_data.guid
        self.reference = self.physical_disk_data.reference

        my_pdr = next((pdr for pdr in self.physical_disk_records if pdr.reference == self.reference), None)
        if not my_pdr:
            raise DDFError(f"Physical disk does not have a physical disk record: {self.reference:#010x}")

        self.type = my_pdr.type
        self.state = my_pdr.state
        self.path_information = my_pdr.path_information
        self.block_size = my_pdr.block_size if my_pdr.block_size != 0xFFFF else self.sector_size

        super().__init__(fh, 0, my_pdr.size * self.block_size)

    def __repr__(self) -> str:
        return f"<DDFPhysicalDisk guid={self.guid} reference={self.reference:#08x} size={self.size}>"


class ControllerData:
    def __init__(self, fh: BinaryIO):
        self.header = _read_section_header(fh, c_ddf.Controller_Data, c_ddf.DDF_CONTROLLER_DATA_SIGNATURE)
        self.guid = self.header.Controller_GUID
        self.type = self.header.Controller_Type

    def __repr__(self) -> str:
        return f"<ControllerData guid={self.guid} type={self.type}>"


class PhysicalDiskData:
    def __init__(self, fh: BinaryIO):
        self.header = _read_section_header(fh, c_ddf.Physical_Disk_Data, c_ddf.DDF_PDD_SIGNATURE)

        self.guid = self.header.PD_GUID
        self.reference = self.header.PD_Reference

    def __repr__(self) -> str:
        return f"<PhysicalDiskData guid={self.guid} reference={self.reference:#x}>"


class PhysicalDiskRecord:
    def __init__(self, fh: BinaryIO):
        self.header = c_ddf.Physical_Disk_Entry(fh)

        self.guid = self.header.PD_GUID
        self.reference = self.header.PD_Reference
        self.type = self.header.PD_Type
        self.state = self.header.PD_State
        self.size = self.header.Configured_Size
        self.path_information = self.header.Path_Information
        self.block_size = self.header.Block_Size

    def __repr__(self) -> str:
        return f"<PhysicalDiskRecord guid={self.guid} reference={self.reference:#x}>"


class VirtualDiskRecord:
    def __init__(self, fh: BinaryIO):
        self.header = c_ddf.Virtual_Disk_Entry(fh)

        self.guid = self.header.VD_GUID
        self.number = self.header.VD_Number
        self.type = self.header.VD_Type
        self.state = self.header.VD_State
        self.init_state = self.header.Init_State
        name = self.header.VD_Name.split(b"\x00")[0]
        self.name = name.decode(encoding="utf-8") if self.type & 0x02 else name.decode(encoding="ascii")

    def __repr__(self) -> str:
        return f"<VirtualDiskRecord guid={self.guid} number={self.number} type={self.type:#x} name={self.name!r}>"


class VirtualDiskConfigurationRecord:
    def __init__(self, fh: BinaryIO, num_entries: int):
        self.header = c_ddf.VD_Configuration_Record(fh)

        self.guid = self.header.VD_GUID
        self.guid_timestamp = ts.from_unix(DECADE + int.from_bytes(self.guid[16:20], "little"))
        self.timestamp = ts.from_unix(DECADE + self.header.Timestamp)
        self.sequence_number = self.header.Sequence_Number
        self.primary_element_count = self.header.Primary_Element_Count
        self.strip_size = self.header.Strip_Size
        self.primary_raid_level = self.header.Primary_RAID_Level
        self.raid_level_qualifier = self.header.RAID_Level_Qualifier
        self.secondary_element_count = self.header.Secondary_Element_Count
        self.secondary_element_seq = self.header.Secondary_Element_Seq
        self.secondary_raid_level = self.header.Secondary_RAID_Level
        self.size = self.header.VD_Size
        self.block_size = self.header.Block_Size
        self.rotate_parity_count = self.header.Rotate_Parity_Count

        self.physical_disk_sequence = c_ddf.uint32[num_entries](fh)[: self.primary_element_count]
        self.starting_block = c_ddf.uint64[num_entries](fh)[: self.primary_element_count]

    def __repr__(self) -> str:
        return f"<VirtualDiskConfigurationRecord guid={self.guid}>"


def _read_section_header(
    fh: BinaryIO,
    structure: type[
        c_ddf.Physical_Disk_Records | c_ddf.Virtual_Disk_Records | c_ddf.Controller_Data | c_ddf.Physical_Disk_Data
    ],
    signature: int,
) -> c_ddf.Physical_Disk_Records | c_ddf.Virtual_Disk_Records | c_ddf.Controller_Data | c_ddf.Physical_Disk_Data:
    obj = structure(fh)
    if obj.Signature != signature:
        raise DDFError(f"Invalid {structure.name} header. Signature: {obj.Signature:#010x}, expected {signature:#x}")
    return obj


def _read_physical_disk_records(fh: BinaryIO, offset: int) -> list[PhysicalDiskRecord]:
    fh.seek(offset)
    header = _read_section_header(fh, c_ddf.Physical_Disk_Records, c_ddf.DDF_PDR_SIGNATURE)
    return [PhysicalDiskRecord(fh) for _ in range(header.Populated_PDEs)]


def _read_virtual_disk_records(fh: BinaryIO, offset: int) -> list[VirtualDiskRecord]:
    fh.seek(offset)
    header = _read_section_header(fh, c_ddf.Virtual_Disk_Records, c_ddf.DDF_VD_RECORD_SIGNATURE)
    return [VirtualDiskRecord(fh) for _ in range(header.Populated_VDEs)]


def _read_virtual_disk_configuration_records(
    fh: BinaryIO, offset: int, count: int, size: int, num_entries: int
) -> list[VirtualDiskConfigurationRecord]:
    result = []
    for _ in range(count):
        fh.seek(offset)
        obj = VirtualDiskConfigurationRecord(fh, num_entries)

        if obj.header.Signature != c_ddf.DDF_VDCR_SIGNATURE:
            continue

        result.append(obj)
        offset += size

    return result


def _convert_raid_layout(prl: int, rlq: int, srl: int, pec: int, sec: int) -> tuple[Level, Layout, int]:
    """Convert DDF RAID layout to a layout we can work with.

    Args:
        prl: The DDF Primary RAID Level.
        rlq: The DDF RAID Level Qualifier.
        srl: The DDF Secondary RAID Level.
        pec: The DDF Primary Element Count
        sec: The DDF Secondary Element Count.
    """
    level = None
    layout = None
    num_disks = pec

    if sec > 1:
        if prl != c_ddf.DDF_VDCR_RAID1 or srl not in (c_ddf.DDF_VDCR_2STRIPED, c_ddf.DDF_VDCR_2SPANNED):
            raise ValueError(f"Unsupported secondary RAID level: {srl}")

        if num_disks == 2 and rlq == c_ddf.DDF_VDCR_RAID1_SIMPLE:
            layout = 0x102  # 2 near copies, 1 far copy
        elif num_disks == 3 and rlq == c_ddf.DDF_VDCR_RAID1_MULTI:
            layout = 0x103  # 3 near copies, 1 far copy

        num_disks *= sec
        level = Level.RAID10

    else:
        if prl == c_ddf.DDF_VDCR_CONCAT:
            level = Level.LINEAR

        elif prl == c_ddf.DDF_VDCR_RAID0:
            if rlq != c_ddf.DDF_VDCR_RAID0_SIMPLE:
                raise ValueError(f"Unsupported DDF RAID0 layout: ({prl}, {rlq})")

            level = Level.RAID0

        elif prl == c_ddf.DDF_VDCR_RAID1:
            if not (
                (rlq == c_ddf.DDF_VDCR_RAID1_SIMPLE and num_disks == 2)
                or (rlq == c_ddf.DDF_VDCR_RAID1_MULTI and num_disks == 3)
            ):
                raise ValueError(f"Unsupported DDF RAID1 layout: ({prl}, {rlq})")

            level = Level.RAID1

        elif prl == c_ddf.DDF_VDCR_RAID1E:
            if rlq == c_ddf.DDF_VDCR_RAID1E_ADJACENT:
                layout = 0x102
            elif rlq == c_ddf.DDF_VDCR_RAID1E_OFFSET:
                layout = 0x201
            else:
                raise ValueError(f"Unsupported DDF RAID1E layout: ({prl}, {rlq})")

            level = Level.RAID10

        elif prl == c_ddf.DDF_VDCR_RAID4:
            if rlq != c_ddf.DDF_VDCR_RAID4_N:
                raise ValueError(f"Unsupported DDF RAID4 layout: ({prl}, {rlq})")

            level = Level.RAID4

        elif prl == c_ddf.DDF_VDCR_RAID5:
            if rlq == c_ddf.DDF_VDCR_RAID5_N_RESTART:
                layout = Layout.LEFT_ASYMMETRIC
            elif rlq == c_ddf.DDF_VDCR_RAID5_0_RESTART:
                layout = Layout.RIGHT_ASYMMETRIC
            elif rlq == c_ddf.DDF_VDCR_RAID5_N_CONTINUE:
                layout = Layout.LEFT_SYMMETRIC
            else:
                raise ValueError(f"Unsupported DDF RAID5 layout: ({prl}, {rlq})")

            level = Level.RAID5

        elif prl == c_ddf.DDF_VDCR_RAID6:
            if rlq == c_ddf.DDF_VDCR_RAID5_N_RESTART:
                layout = Layout.ROTATING_N_RESTART
            elif rlq == c_ddf.DDF_VDCR_RAID6_0_RESTART:
                layout = Layout.ROTATING_ZERO_RESTART
            elif rlq == c_ddf.DDF_VDCR_RAID5_N_CONTINUE:
                layout = Layout.ROTATING_N_CONTINUE
            else:
                raise ValueError(f"Unsupported DDF RAID6 layout: ({prl}, {rlq})")

            level = Level.RAID6
        else:
            raise ValueError(f"Unsupported DDF RAID layout: ({prl}, {rlq})")

    return level, layout, num_disks



from __future__ import annotations

from dissect.cstruct import cstruct

dm_def = """
/* DM BTree */
enum node_flags {
    INTERNAL_NODE = 1,
    LEAF_NODE = 1 << 1
};

/*
 * Every btree node begins with this structure.  Make sure it's a multiple
 * of 8-bytes in size, otherwise the 64bit keys will be mis-aligned.
 */
struct node_header {
    uint32  csum;
    uint32  flags;
    uint64  blocknr;                /* Block this node is supposed to live in. */

    uint32  nr_entries;
    uint32  max_entries;
    uint32  value_size;
    uint32  padding;
};

struct btree_node {
    struct node_header header;
    // uint64 keys[];
};

/* Thin */
#define THIN_SUPERBLOCK_MAGIC 27022010
#define THIN_SUPERBLOCK_LOCATION 0
#define THIN_VERSION 2
#define SECTOR_TO_BLOCK_SHIFT 3

/* This should be plenty */
#define SPACE_MAP_ROOT_SIZE 128

/*
 * Little endian on-disk superblock and device details.
 */
struct thin_disk_superblock {
    uint32  csum;                   /* Checksum of superblock except for this field. */
    uint32  flags;
    uint64  blocknr;                /* This block number, dm_block_t. */

    uint8   uuid[16];
    uint64  magic;
    uint32  version;
    uint32  time;

    uint64  trans_id;

    /*
     * Root held by userspace transactions.
     */
    uint64  held_root;

    uint8   data_space_map_root[SPACE_MAP_ROOT_SIZE];
    uint8   metadata_space_map_root[SPACE_MAP_ROOT_SIZE];

    /*
     * 2-level btree mapping (dev_id, (dev block, time)) -> data block
     */
    uint64  data_mapping_root;

    /*
     * Device detail root mapping dev_id -> device_details
     */
    uint64  device_details_root;

    uint32  data_block_size;        /* In 512-byte sectors. */

    uint32  metadata_block_size;    /* In 512-byte sectors. */
    uint64  metadata_nr_blocks;

    uint32  compat_flags;
    uint32  compat_ro_flags;
    uint32  incompat_flags;
};

struct disk_device_details {
    uint64 mapped_blocks;
    uint64 transaction_id;          /* When created. */
    uint32 creation_time;
    uint32 snapshotted_time;
};
"""

c_dm = cstruct().load(dm_def)

SECTOR_SIZE = 512



# Reference:
# - https://github.com/freebsd/freebsd-src/blob/main/sys/geom/part/g_part_bsd.c
from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.cstruct import cstruct

from dissect.volume.disk.partition import Partition
from dissect.volume.exceptions import DiskError

if TYPE_CHECKING:
    from collections.abc import Iterator

bsd_def = """
/* The disk magic number */
#define BSD_MAGIC           0x82564557

#define BSD_NPARTS_MIN      8
#define BSD_NPARTS_MAX      20

#define DISKMAGIC64         0xc4464c59
#define MAXPARTITIONS64     16
#define RESPARTITIONS64     32

/* Size of bootblock area in sector-size neutral bytes */
#define BSD_BOOTBLOCK_SIZE  8192

/* partition containing whole disk */
#define BSD_PART_RAW        2

/* partition normally containing swap */
#define BSD_PART_SWAP       1

/* Drive-type specific data size (in number of 32-bit inegrals) */
#define BSD_NDRIVEDATA      5

/* Number of spare 32-bit integrals following drive-type data */
#define BSD_NSPARE          5

struct partition {                              /* the partition table */
    uint32_t    p_size;                         /* number of sectors in partition */
    uint32_t    p_offset;                       /* starting sector */
    uint32_t    p_fsize;                        /* filesystem basic fragment size */
    uint8_t     p_fstype;                       /* filesystem type, see below */
    uint8_t     p_frag;                         /* filesystem fragments per block */
    uint16_t    p_cpg;                          /* filesystem cylinders per group */
};

struct partition64 {                            /* the partition table */
    uint64_t    p_boffset;                      /* slice relative offset, in bytes */
    uint64_t    p_bsize;                        /* size of partition, in bytes */
    uint8_t     p_fstype;
    uint8_t     p_unused01;                     /* reserved, must be 0 */
    uint8_t     p_unused02;                     /* reserved, must be 0 */
    uint8_t     p_unused03;                     /* reserved, must be 0 */
    uint32_t    p_unused04;                     /* reserved, must be 0 */
    uint32_t    p_unused05;                     /* reserved, must be 0 */
    uint32_t    p_unused06;                     /* reserved, must be 0 */
    char        p_type_uuid[16];                /* mount type as UUID */
    char        p_stor_uuid[16];                /* unique uuid for storage */
};

struct disklabel {
    uint32_t    d_magic;                        /* the magic number */
    uint16_t    d_type;                         /* drive type */
    uint16_t    d_subtype;                      /* controller/d_type specific */
    char        d_typename[16];                 /* type name, e.g. "eagle" */

    char        d_packname[16];                 /* pack identifier */

    /* disk geometry: */
    uint32_t    d_secsize;                      /* # of bytes per sector */
    uint32_t    d_nsectors;                     /* # of data sectors per track */
    uint32_t    d_ntracks;                      /* # of tracks per cylinder */
    uint32_t    d_ncylinders;                   /* # of data cylinders per unit */
    uint32_t    d_secpercyl;                    /* # of data sectors per cylinder */
    uint32_t    d_secperunit;                   /* # of data sectors per unit */

    /*
     * Spares (bad sector replacements) below are not counted in
     * d_nsectors or d_secpercyl.  Spare sectors are assumed to
     * be physical sectors which occupy space at the end of each
     * track and/or cylinder.
     */
    uint16_t    d_sparespertrack;               /* # of spare sectors per track */
    uint16_t    d_sparespercyl;                 /* # of spare sectors per cylinder */
    /*
     * Alternate cylinders include maintenance, replacement, configuration
     * description areas, etc.
     */
    uint32_t    d_acylinders;                   /* # of alt. cylinders per unit */

    /* hardware characteristics: */
    /*
     * d_interleave, d_trackskew and d_cylskew describe perturbations
     * in the media format used to compensate for a slow controller.
     * Interleave is physical sector interleave, set up by the
     * formatter or controller when formatting.  When interleaving is
     * in use, logically adjacent sectors are not physically
     * contiguous, but instead are separated by some number of
     * sectors.  It is specified as the ratio of physical sectors
     * traversed per logical sector.  Thus an interleave of 1:1
     * implies contiguous layout, while 2:1 implies that logical
     * sector 0 is separated by one sector from logical sector 1.
     * d_trackskew is the offset of sector 0 on track N relative to
     * sector 0 on track N-1 on the same cylinder.  Finally, d_cylskew
     * is the offset of sector 0 on cylinder N relative to sector 0
     * on cylinder N-1.
     */
    uint16_t    d_rpm;                          /* rotational speed */
    uint16_t    d_interleave;                   /* hardware sector interleave */
    uint16_t    d_trackskew;                    /* sector 0 skew, per track */
    uint16_t    d_cylskew;                      /* sector 0 skew, per cylinder */
    uint32_t    d_headswitch;                   /* head switch time, usec */
    uint32_t    d_trkseek;                      /* track-to-track seek, usec */
    uint32_t    d_flags;                        /* generic flags */
    uint32_t    d_drivedata[BSD_NDRIVEDATA];    /* drive-type specific data */
    uint32_t    d_spare[BSD_NSPARE];            /* reserved for future use */
    uint32_t    d_magic2;                       /* the magic number (again) */
    uint16_t    d_checksum;                     /* xor of data incl. partitions */

    /* filesystem and partition information: */
    uint16_t    d_npartitions;                  /* number of partitions in following */
    uint32_t    d_bbsize;                       /* size of boot area at sn0, bytes */
    uint32_t    d_sbsize;                       /* max size of fs superblock, bytes */
    // partition   d_partitions[BSD_NPARTS_MIN];   /* actually may be more */
};

struct disklabel64 {
    uint32_t    d_magic;                        /* the magic number */
    uint32_t    d_crc;                          /* crc32() d_magic through last part */
    uint32_t    d_align;                        /* partition alignment requirement */
    uint32_t    d_npartitions;                  /* number of partitions */
    char        d_stor_uuid[16];                /* unique uuid for label */

    uint64_t    d_total_size;                   /* total size incl everything (bytes) */
    uint64_t    d_bbase;                        /* boot area base offset (bytes) */
                                                /* boot area is pbase - bbase */
    uint64_t    d_pbase;                        /* first allocatable offset (bytes) */
    uint64_t    d_pstop;                        /* last allocatable offset+1 (bytes) */
    uint64_t    d_abase;                        /* location of backup copy if not 0 */

    char        d_packname[64];
    char        d_reserved[64];

    /*
     * Note: offsets are relative to the base of the slice, NOT to
     * d_pbase.  Unlike 32 bit disklabels the on-disk format for
     * a 64 bit disklabel remains slice-relative.
     *
     * An uninitialized partition has a p_boffset and p_bsize of 0.
     *
     * If p_fstype is not supported for a live partition it is set
     * to FS_OTHER.  This is typically the case when the filesystem
     * is identified by its uuid.
     */
    //partition64 d_partitions[MAXPARTITIONS64]; /* actually may be more */
};

/* d_type values: */
#define DTYPE_SMD           1                   /* SMD, XSMD; VAX hp/up */
#define DTYPE_MSCP          2                   /* MSCP */
#define DTYPE_DEC           3                   /* other DEC (rk, rl) */
#define DTYPE_SCSI          4                   /* SCSI */
#define DTYPE_ESDI          5                   /* ESDI interface */
#define DTYPE_ST506         6                   /* ST506 etc. */
#define DTYPE_HPIB          7                   /* CS/80 on HP-IB */
#define DTYPE_HPFL          8                   /* HP Fiber-link */
#define DTYPE_FLOPPY        10                  /* floppy */
#define DTYPE_CCD           11                  /* concatenated disk */
#define DTYPE_VINUM         12                  /* vinum volume */
#define DTYPE_DOC2K         13                  /* Msys DiskOnChip */
#define DTYPE_RAID          14                  /* CMU RAIDFrame */
#define DTYPE_JFS2          16                  /* IBM JFS 2 */

/*
 * Filesystem type and version.
 * Used to interpret other filesystem-specific
 * per-partition information.
 */
#define FS_UNUSED           0                   /* unused */
#define FS_SWAP             1                   /* swap */
#define FS_V6               2                   /* Sixth Edition */
#define FS_V7               3                   /* Seventh Edition */
#define FS_SYSV             4                   /* System V */
#define FS_V71K             5                   /* V7 with 1K blocks (4.1, 2.9) */
#define FS_V8               6                   /* Eighth Edition, 4K blocks */
#define FS_BSDFFS           7                   /* 4.2BSD fast filesystem */
#define FS_MSDOS            8                   /* MSDOS filesystem */
#define FS_BSDLFS           9                   /* 4.4BSD log-structured filesystem */
#define FS_OTHER            10                  /* in use, but unknown/unsupported */
#define FS_HPFS             11                  /* OS/2 high-performance filesystem */
#define FS_ISO9660          12                  /* ISO 9660, normally CD-ROM */
#define FS_BOOT             13                  /* partition contains bootstrap */
#define FS_VINUM            14                  /* Vinum drive */
#define FS_RAID             15                  /* RAIDFrame drive */
#define FS_FILECORE         16                  /* Acorn Filecore Filing System */
#define FS_EXT2FS           17                  /* ext2fs */
#define FS_NTFS             18                  /* Windows/NT file system */
#define FS_CCD              20                  /* concatenated disk component */
#define FS_JFS2             21                  /* IBM JFS2 */
#define FS_HAMMER           22                  /* DragonFlyBSD Hammer FS */
#define FS_HAMMER2          23                  /* DragonFlyBSD Hammer2 FS */
#define FS_UDF              24                  /* UDF */
#define FS_EFS              26                  /* SGI's Extent File system */
#define FS_ZFS              27                  /* Sun's ZFS */
#define FS_NANDFS           30                  /* FreeBSD nandfs (NiLFS derived) */

/*
 * flags shared by various drives:
 */
#define D_REMOVABLE         0x01                /* removable media */
#define D_ECC               0x02                /* supports ECC */
#define D_BADSECT           0x04                /* supports bad sector forw. */
#define D_RAMDISK           0x08                /* disk emulator */
#define D_CHAIN             0x10                /* can do back-back transfers */
"""

c_bsd = cstruct().load(bsd_def)

DTYPE_NAMES = {
    0: "unknown",
    c_bsd.DTYPE_SMD: "SMD",
    c_bsd.DTYPE_MSCP: "MSCP",
    c_bsd.DTYPE_DEC: "old DEC",
    c_bsd.DTYPE_SCSI: "SCSI",
    c_bsd.DTYPE_ESDI: "ESDI",
    c_bsd.DTYPE_ST506: "ST506",
    c_bsd.DTYPE_HPIB: "HP-IB",
    c_bsd.DTYPE_HPFL: "HP-FL",
    9: "type 9",
    c_bsd.DTYPE_FLOPPY: "floppy",
    c_bsd.DTYPE_CCD: "CCD",
    c_bsd.DTYPE_VINUM: "Vinum",
    c_bsd.DTYPE_DOC2K: "DOC2K",
    c_bsd.DTYPE_RAID: "Raid",
    c_bsd.DTYPE_JFS2: "jfs",
}

FS_NAMES = {
    c_bsd.FS_UNUSED: "unused",
    c_bsd.FS_SWAP: "swap",
    c_bsd.FS_V6: "Version 6",
    c_bsd.FS_V7: "Version 7",
    c_bsd.FS_SYSV: "System V",
    c_bsd.FS_V71K: "4.1BSD",
    c_bsd.FS_V8: "Eighth Edition",
    c_bsd.FS_BSDFFS: "4.2BSD",
    c_bsd.FS_MSDOS: "MSDOS",
    c_bsd.FS_BSDLFS: "4.4LFS",
    c_bsd.FS_OTHER: "unknown",
    c_bsd.FS_HPFS: "HPFS",
    c_bsd.FS_ISO9660: "ISO9660",
    c_bsd.FS_BOOT: "boot",
    c_bsd.FS_VINUM: "vinum",
    c_bsd.FS_RAID: "raid",
    c_bsd.FS_FILECORE: "Filecore",
    c_bsd.FS_EXT2FS: "EXT2FS",
    c_bsd.FS_NTFS: "NTFS",
    c_bsd.FS_CCD: "ccd",
    c_bsd.FS_JFS2: "jfs",
    c_bsd.FS_HAMMER: "HAMMER",
    c_bsd.FS_HAMMER2: "HAMMER2",
    c_bsd.FS_UDF: "UDF",
    c_bsd.FS_EFS: "EFS",
    c_bsd.FS_ZFS: "ZFS",
    c_bsd.FS_NANDFS: "nandfs",
}


class BSD:
    """BSD disklabel."""

    TYPES = (
        # MBR FreeBSD
        0xA5,
        # MBR OpenBSD
        0xA6,
        # MBR NetBSD
        0xA9,
        # MBR DragonFlyBSD
        0x6C,
        # GPT DragonFlyBSD disklabel32
        UUID("9D087404-1CA5-11DC-8817-01301BB8A9F5"),
        # GPT DragonFlyBSD disklabel64
        UUID("3D48CE54-1D16-11DC-8696-01301BB8A9F5"),
        # GPT FreeBSD disklabel
        UUID("516E7CB4-6ECF-11D6-8FF8-00022D09712B"),
    )

    def __init__(self, fh: BinaryIO, sector_size: int = 512):
        self.fh = fh
        self.sector_size = sector_size

        # BSD disklabel is usually part of another partition scheme such as MBR or GPT.
        # The actual BSD disklabel info always starts at the second sector.
        offset = fh.seek(sector_size, io.SEEK_CUR)
        data = fh.read(512)
        magic = c_bsd.uint32(data)

        if magic == c_bsd.BSD_MAGIC:
            # disklabel
            self.disklabel = c_bsd.disklabel(data)
            self.type = 32

            if (self.disklabel.d_magic, self.disklabel.d_magic2) != (c_bsd.BSD_MAGIC, c_bsd.BSD_MAGIC):
                raise DiskError(
                    f"Invalid BSD disklabel magic, expected {c_bsd.BSD_MAGIC:#x}, "
                    f"got ({self.disklabel.d_magic:#x}, {self.disklabel.d_magic2:#x})."
                )
        elif magic == c_bsd.DISKMAGIC64:
            # disklabel64
            self.disklabel = c_bsd.disklabel64(data)
            self.type = 64
        else:
            raise DiskError(
                f"Invalid BSD disklabel magic, expected {c_bsd.BSD_MAGIC:#x} or {c_bsd.DISKMAGIC64:#x}, got {magic:#x}."
            )

        self._partitions_offset = offset + len(self.disklabel)
        self.partitions = list(self._partitions())

    def _partitions(self) -> Iterator[Partition]:
        table_offset = 0
        if self.type == 32:
            # Get the table offset first
            self.fh.seek(self._partitions_offset + (c_bsd.BSD_PART_RAW * len(c_bsd.partition)))
            table_offset = c_bsd.partition(self.fh).p_offset

        self.fh.seek(self._partitions_offset)
        for i in range(self.disklabel.d_npartitions):
            if self.type == 32:
                partition = c_bsd.partition(self.fh)
                if i == c_bsd.BSD_PART_RAW:
                    # Skip internal partition
                    continue

                if partition.p_size == 0 or partition.p_fstype == 0:
                    continue

                if partition.p_offset < table_offset:
                    continue

                offset = (partition.p_offset - table_offset) * self.sector_size
                size = partition.p_size * self.sector_size
                guid = None
            elif self.type == 64:
                partition = c_bsd.partition64(self.fh)
                if i == c_bsd.BSD_PART_RAW:
                    # Skip internal partition
                    continue

                if (partition.p_boffset == 0 and partition.p_bsize == 0) or partition.p_fstype == 0:
                    continue

                offset = partition.p_boffset
                size = partition.p_bsize
                guid = UUID(bytes_le=partition.p_stor_uuid)

            yield Partition(
                disk=self,
                number=i + 1,  # partitions are 1-indexed
                offset=offset,
                size=size,
                type=partition.p_fstype,
                guid=guid,
                type_str=FS_NAMES.get(partition.p_fstype, "?"),
                raw=partition,
            )



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_ddf(__cs__.cstruct):
    DDF_HEADER_SIGNATURE: Literal[3725712913] = ...
    DDF_HEADER_CLOSED: Literal[0] = ...
    DDF_HEADER_OPENED_MASK: Literal[15] = ...
    DDF_HEADER_OPEN_ANCHOR: Literal[255] = ...
    DDF_HEADER_ANCHOR: Literal[0] = ...
    DDF_HEADER_PRIMARY: Literal[1] = ...
    DDF_HEADER_SECONDARY: Literal[2] = ...
    DDF_CONTROLLER_DATA_SIGNATURE: Literal[2903576849] = ...
    DDF_PDR_SIGNATURE: Literal[572662306] = ...
    DDF_PDE_GUID_FORCE: Literal[1] = ...
    DDF_PDE_PARTICIPATING: Literal[2] = ...
    DDF_PDE_GLOBAL_SPARE: Literal[4] = ...
    DDF_PDE_CONFIG_SPARE: Literal[8] = ...
    DDF_PDE_FOREIGN: Literal[16] = ...
    DDF_PDE_LEGACY: Literal[32] = ...
    DDF_PDE_TYPE_MASK: Literal[61440] = ...
    DDF_PDE_UNKNOWN: Literal[0] = ...
    DDF_PDE_SCSI: Literal[4096] = ...
    DDF_PDE_SAS: Literal[8192] = ...
    DDF_PDE_SATA: Literal[12288] = ...
    DDF_PDE_FC: Literal[16384] = ...
    DDF_PDE_ONLINE: Literal[1] = ...
    DDF_PDE_FAILED: Literal[2] = ...
    DDF_PDE_REBUILD: Literal[4] = ...
    DDF_PDE_TRANSITION: Literal[8] = ...
    DDF_PDE_PFA: Literal[16] = ...
    DDF_PDE_UNRECOVERED: Literal[32] = ...
    DDF_PDE_MISSING: Literal[64] = ...
    DDF_VD_RECORD_SIGNATURE: Literal[3722304989] = ...
    DDF_VDE_SHARED: Literal[1] = ...
    DDF_VDE_ENFORCE_GROUP: Literal[2] = ...
    DDF_VDE_UNICODE_NAME: Literal[4] = ...
    DDF_VDE_OWNER_ID_VALID: Literal[8] = ...
    DDF_VDE_OPTIMAL: Literal[0] = ...
    DDF_VDE_DEGRADED: Literal[1] = ...
    DDF_VDE_DELETED: Literal[2] = ...
    DDF_VDE_MISSING: Literal[3] = ...
    DDF_VDE_FAILED: Literal[4] = ...
    DDF_VDE_PARTIAL: Literal[5] = ...
    DDF_VDE_OFFLINE: Literal[6] = ...
    DDF_VDE_STATE_MASK: Literal[7] = ...
    DDF_VDE_MORPH: Literal[8] = ...
    DDF_VDE_DIRTY: Literal[16] = ...
    DDF_VDE_UNINTIALIZED: Literal[0] = ...
    DDF_VDE_INIT_QUICK: Literal[1] = ...
    DDF_VDE_INIT_FULL: Literal[2] = ...
    DDF_VDE_INIT_MASK: Literal[3] = ...
    DDF_VDE_UACCESS_RW: Literal[0] = ...
    DDF_VDE_UACCESS_RO: Literal[128] = ...
    DDF_VDE_UACCESS_BLOCKED: Literal[192] = ...
    DDF_VDE_UACCESS_MASK: Literal[192] = ...
    DDF_VDCR_SIGNATURE: Literal[4008636142] = ...
    DDF_VDCR_RAID0: Literal[0] = ...
    DDF_VDCR_RAID1: Literal[1] = ...
    DDF_VDCR_RAID3: Literal[3] = ...
    DDF_VDCR_RAID4: Literal[4] = ...
    DDF_VDCR_RAID5: Literal[5] = ...
    DDF_VDCR_RAID6: Literal[6] = ...
    DDF_VDCR_RAID1E: Literal[17] = ...
    DDF_VDCR_SINGLE: Literal[15] = ...
    DDF_VDCR_CONCAT: Literal[31] = ...
    DDF_VDCR_RAID5E: Literal[21] = ...
    DDF_VDCR_RAID5EE: Literal[37] = ...
    DDF_VDCR_RAID0_SIMPLE: Literal[0] = ...
    DDF_VDCR_RAID1_SIMPLE: Literal[0] = ...
    DDF_VDCR_RAID1_MULTI: Literal[1] = ...
    DDF_VDCR_RAID3_0: Literal[0] = ...
    DDF_VDCR_RAID3_N: Literal[1] = ...
    DDF_VDCR_RAID4_0: Literal[0] = ...
    DDF_VDCR_RAID4_N: Literal[1] = ...
    DDF_VDCR_RAID5_0_RESTART: Literal[0] = ...
    DDF_VDCR_RAID6_0_RESTART: Literal[1] = ...
    DDF_VDCR_RAID5_N_RESTART: Literal[2] = ...
    DDF_VDCR_RAID5_N_CONTINUE: Literal[3] = ...
    DDF_VDCR_RAID1E_ADJACENT: Literal[0] = ...
    DDF_VDCR_RAID1E_OFFSET: Literal[1] = ...
    DDF_VDCR_2STRIPED: Literal[0] = ...
    DDF_VDCR_2MIRRORED: Literal[1] = ...
    DDF_VDCR_2CONCAT: Literal[2] = ...
    DDF_VDCR_2SPANNED: Literal[3] = ...
    DDF_VDCR_CACHE_WB: Literal[1] = ...
    DDF_VDCR_CACHE_WB_ADAPTIVE: Literal[2] = ...
    DDF_VDCR_CACHE_RA: Literal[4] = ...
    DDF_VDCR_CACHE_RA_ADAPTIVE: Literal[8] = ...
    DDF_VDCR_CACHE_WCACHE_NOBATTERY: Literal[16] = ...
    DDF_VDCR_CACHE_WCACHE_ALLOW: Literal[32] = ...
    DDF_VDCR_CACHE_RCACHE_ALLOW: Literal[64] = ...
    DDF_VDCR_CACHE_VENDOR: Literal[128] = ...
    DDF_VUCR_SIGNATURE: Literal[2290649224] = ...
    DDF_SA_SIGNATURE: Literal[1431655765] = ...
    DDF_SAR_TYPE_DEDICATED: Literal[1] = ...
    DDF_SAR_TYPE_REVERTIBLE: Literal[2] = ...
    DDF_SAR_TYPE_ACTIVE: Literal[4] = ...
    DDF_SAR_TYPE_ENCL_AFFINITY: Literal[8] = ...
    DDF_PDD_SIGNATURE: Literal[858993459] = ...
    DDF_PDD_FORCED_REF: Literal[1] = ...
    DDF_PDD_FORCED_GUID: Literal[1] = ...
    DDF_BBML_SIGNATURE: Literal[2880286988] = ...
    DDF_VENDOR_LOG_SIGNATURE: Literal[31190768] = ...
    class DDF_Header(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        DDF_Header_GUID: __cs__.CharArray
        DDF_rev: __cs__.CharArray
        Sequence_Number: _c_ddf.uint32
        TimeStamp: _c_ddf.uint32
        Open_Flag: _c_ddf.uint8
        Foreign_Flag: _c_ddf.uint8
        Disk_Grouping: _c_ddf.uint8
        Reserved1: __cs__.CharArray
        Header_Ext: __cs__.Array[_c_ddf.uint8]
        Primary_Header_LBA: _c_ddf.uint64
        Secondary_Header_LBA: _c_ddf.uint64
        Header_Type: _c_ddf.uint8
        Reserved2: __cs__.CharArray
        Workspace_Length: _c_ddf.uint32
        Workspace_LBA: _c_ddf.uint64
        Max_PD_Entries: _c_ddf.uint16
        Max_VD_Entries: _c_ddf.uint16
        Max_Partitions: _c_ddf.uint16
        Configuration_Record_Length: _c_ddf.uint16
        Max_Primary_Element_Entries: _c_ddf.uint16
        Max_Mapped_Block_Entries: _c_ddf.uint32
        Reserved3: __cs__.CharArray
        Controller_Data_Section: _c_ddf.uint32
        Controller_Data_Section_Length: _c_ddf.uint32
        Physical_Disk_Records_Section: _c_ddf.uint32
        Physical_Disk_Records_Section_Length: _c_ddf.uint32
        Virtual_Disk_Records_Section: _c_ddf.uint32
        Virtual_Disk_Records_Section_Length: _c_ddf.uint32
        Configuration_Records_Section: _c_ddf.uint32
        Configuration_Records_Section_Length: _c_ddf.uint32
        Physical_Disk_Data_Section: _c_ddf.uint32
        Physical_Disk_Data_Section_Length: _c_ddf.uint32
        BBM_Log_Section: _c_ddf.uint32
        BBM_Log_Section_Length: _c_ddf.uint32
        Diagnostic_Space_Section: _c_ddf.uint32
        Diagnostic_Space_Section_Length: _c_ddf.uint32
        Vendor_Specific_Logs_Section: _c_ddf.uint32
        Vendor_Specific_Logs_Section_Length: _c_ddf.uint32
        Reserved4: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            DDF_Header_GUID: __cs__.CharArray | None = ...,
            DDF_rev: __cs__.CharArray | None = ...,
            Sequence_Number: _c_ddf.uint32 | None = ...,
            TimeStamp: _c_ddf.uint32 | None = ...,
            Open_Flag: _c_ddf.uint8 | None = ...,
            Foreign_Flag: _c_ddf.uint8 | None = ...,
            Disk_Grouping: _c_ddf.uint8 | None = ...,
            Reserved1: __cs__.CharArray | None = ...,
            Header_Ext: __cs__.Array[_c_ddf.uint8] | None = ...,
            Primary_Header_LBA: _c_ddf.uint64 | None = ...,
            Secondary_Header_LBA: _c_ddf.uint64 | None = ...,
            Header_Type: _c_ddf.uint8 | None = ...,
            Reserved2: __cs__.CharArray | None = ...,
            Workspace_Length: _c_ddf.uint32 | None = ...,
            Workspace_LBA: _c_ddf.uint64 | None = ...,
            Max_PD_Entries: _c_ddf.uint16 | None = ...,
            Max_VD_Entries: _c_ddf.uint16 | None = ...,
            Max_Partitions: _c_ddf.uint16 | None = ...,
            Configuration_Record_Length: _c_ddf.uint16 | None = ...,
            Max_Primary_Element_Entries: _c_ddf.uint16 | None = ...,
            Max_Mapped_Block_Entries: _c_ddf.uint32 | None = ...,
            Reserved3: __cs__.CharArray | None = ...,
            Controller_Data_Section: _c_ddf.uint32 | None = ...,
            Controller_Data_Section_Length: _c_ddf.uint32 | None = ...,
            Physical_Disk_Records_Section: _c_ddf.uint32 | None = ...,
            Physical_Disk_Records_Section_Length: _c_ddf.uint32 | None = ...,
            Virtual_Disk_Records_Section: _c_ddf.uint32 | None = ...,
            Virtual_Disk_Records_Section_Length: _c_ddf.uint32 | None = ...,
            Configuration_Records_Section: _c_ddf.uint32 | None = ...,
            Configuration_Records_Section_Length: _c_ddf.uint32 | None = ...,
            Physical_Disk_Data_Section: _c_ddf.uint32 | None = ...,
            Physical_Disk_Data_Section_Length: _c_ddf.uint32 | None = ...,
            BBM_Log_Section: _c_ddf.uint32 | None = ...,
            BBM_Log_Section_Length: _c_ddf.uint32 | None = ...,
            Diagnostic_Space_Section: _c_ddf.uint32 | None = ...,
            Diagnostic_Space_Section_Length: _c_ddf.uint32 | None = ...,
            Vendor_Specific_Logs_Section: _c_ddf.uint32 | None = ...,
            Vendor_Specific_Logs_Section_Length: _c_ddf.uint32 | None = ...,
            Reserved4: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Controller_Data(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Controller_GUID: __cs__.CharArray
        Controller_Type: __cs__.CharArray
        Product_ID: __cs__.CharArray
        Reserved: __cs__.CharArray
        Vendor_Unique_Controller_Data: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Controller_GUID: __cs__.CharArray | None = ...,
            Controller_Type: __cs__.CharArray | None = ...,
            Product_ID: __cs__.CharArray | None = ...,
            Reserved: __cs__.CharArray | None = ...,
            Vendor_Unique_Controller_Data: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Physical_Disk_Entry(__cs__.Structure):
        PD_GUID: __cs__.CharArray
        PD_Reference: _c_ddf.uint32
        PD_Type: _c_ddf.uint16
        PD_State: _c_ddf.uint16
        Configured_Size: _c_ddf.uint64
        Path_Information: __cs__.CharArray
        Block_Size: _c_ddf.uint16
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            PD_GUID: __cs__.CharArray | None = ...,
            PD_Reference: _c_ddf.uint32 | None = ...,
            PD_Type: _c_ddf.uint16 | None = ...,
            PD_State: _c_ddf.uint16 | None = ...,
            Configured_Size: _c_ddf.uint64 | None = ...,
            Path_Information: __cs__.CharArray | None = ...,
            Block_Size: _c_ddf.uint16 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Physical_Disk_Records(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Populated_PDEs: _c_ddf.uint16
        Max_PDE_Supported: _c_ddf.uint16
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Populated_PDEs: _c_ddf.uint16 | None = ...,
            Max_PDE_Supported: _c_ddf.uint16 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Virtual_Disk_Entry(__cs__.Structure):
        VD_GUID: __cs__.CharArray
        VD_Number: _c_ddf.uint16
        Reserved1: __cs__.CharArray
        VD_Type: _c_ddf.uint32
        VD_State: _c_ddf.uint8
        Init_State: _c_ddf.uint8
        Partially_Optimal_Drive_Failures_Remaining: _c_ddf.uint8
        Reserved2: __cs__.CharArray
        VD_Name: __cs__.CharArray
        @overload
        def __init__(
            self,
            VD_GUID: __cs__.CharArray | None = ...,
            VD_Number: _c_ddf.uint16 | None = ...,
            Reserved1: __cs__.CharArray | None = ...,
            VD_Type: _c_ddf.uint32 | None = ...,
            VD_State: _c_ddf.uint8 | None = ...,
            Init_State: _c_ddf.uint8 | None = ...,
            Partially_Optimal_Drive_Failures_Remaining: _c_ddf.uint8 | None = ...,
            Reserved2: __cs__.CharArray | None = ...,
            VD_Name: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Virtual_Disk_Records(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Populated_VDEs: _c_ddf.uint16
        Max_VDE_Supported: _c_ddf.uint16
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Populated_VDEs: _c_ddf.uint16 | None = ...,
            Max_VDE_Supported: _c_ddf.uint16 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class VD_Configuration_Record(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        VD_GUID: __cs__.CharArray
        Timestamp: _c_ddf.uint32
        Sequence_Number: _c_ddf.uint32
        Reserved1: __cs__.CharArray
        Primary_Element_Count: _c_ddf.uint16
        Strip_Size: _c_ddf.uint8
        Primary_RAID_Level: _c_ddf.uint8
        RAID_Level_Qualifier: _c_ddf.uint8
        Secondary_Element_Count: _c_ddf.uint8
        Secondary_Element_Seq: _c_ddf.uint8
        Secondary_RAID_Level: _c_ddf.uint8
        Block_Count: _c_ddf.uint64
        VD_Size: _c_ddf.uint64
        Block_Size: _c_ddf.uint16
        Rotate_Parity_Count: _c_ddf.uint8
        Reserved2: __cs__.CharArray
        Associated_Spares: __cs__.CharArray
        Cache_Policies_And_Parameters: _c_ddf.uint64
        BG_Rate: _c_ddf.uint8
        Reserved3: __cs__.CharArray
        MDF_Parity_Disks: _c_ddf.uint8
        MDF_Parity_Generator_Polynomial: _c_ddf.uint16
        Reserved4: __cs__.CharArray
        MDF_Constant_Generation_Method: _c_ddf.uint8
        Reserved5: __cs__.CharArray
        Reserved6: __cs__.CharArray
        V0: __cs__.CharArray
        V1: __cs__.CharArray
        V2: __cs__.CharArray
        V3: __cs__.CharArray
        Vendor_Specific_Scratch_Space: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            VD_GUID: __cs__.CharArray | None = ...,
            Timestamp: _c_ddf.uint32 | None = ...,
            Sequence_Number: _c_ddf.uint32 | None = ...,
            Reserved1: __cs__.CharArray | None = ...,
            Primary_Element_Count: _c_ddf.uint16 | None = ...,
            Strip_Size: _c_ddf.uint8 | None = ...,
            Primary_RAID_Level: _c_ddf.uint8 | None = ...,
            RAID_Level_Qualifier: _c_ddf.uint8 | None = ...,
            Secondary_Element_Count: _c_ddf.uint8 | None = ...,
            Secondary_Element_Seq: _c_ddf.uint8 | None = ...,
            Secondary_RAID_Level: _c_ddf.uint8 | None = ...,
            Block_Count: _c_ddf.uint64 | None = ...,
            VD_Size: _c_ddf.uint64 | None = ...,
            Block_Size: _c_ddf.uint16 | None = ...,
            Rotate_Parity_Count: _c_ddf.uint8 | None = ...,
            Reserved2: __cs__.CharArray | None = ...,
            Associated_Spares: __cs__.CharArray | None = ...,
            Cache_Policies_And_Parameters: _c_ddf.uint64 | None = ...,
            BG_Rate: _c_ddf.uint8 | None = ...,
            Reserved3: __cs__.CharArray | None = ...,
            MDF_Parity_Disks: _c_ddf.uint8 | None = ...,
            MDF_Parity_Generator_Polynomial: _c_ddf.uint16 | None = ...,
            Reserved4: __cs__.CharArray | None = ...,
            MDF_Constant_Generation_Method: _c_ddf.uint8 | None = ...,
            Reserved5: __cs__.CharArray | None = ...,
            Reserved6: __cs__.CharArray | None = ...,
            V0: __cs__.CharArray | None = ...,
            V1: __cs__.CharArray | None = ...,
            V2: __cs__.CharArray | None = ...,
            V3: __cs__.CharArray | None = ...,
            Vendor_Specific_Scratch_Space: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class VU_Configuration_Record(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        VD_GUID: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            VD_GUID: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Spare_Assignment_Entry(__cs__.Structure):
        VD_GUID: __cs__.CharArray
        Secondary_Element: _c_ddf.uint16
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            VD_GUID: __cs__.CharArray | None = ...,
            Secondary_Element: _c_ddf.uint16 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Spare_Assignment_Record(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Timestamp: _c_ddf.uint32
        Reserved1: __cs__.CharArray
        Spare_Type: _c_ddf.uint8
        Populated_SAEs: _c_ddf.uint16
        Max_SAE_Supported: _c_ddf.uint16
        Reserved2: __cs__.CharArray
        class Spare_Assignment_Entry(__cs__.Structure):
            VD_GUID: __cs__.CharArray
            Secondary_Element: _c_ddf.uint16
            Reserved: __cs__.CharArray
            @overload
            def __init__(
                self,
                VD_GUID: __cs__.CharArray | None = ...,
                Secondary_Element: _c_ddf.uint16 | None = ...,
                Reserved: __cs__.CharArray | None = ...,
            ): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        Spare_Assignment_Entries: __cs__.Array[Spare_Assignment_Entry]
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Timestamp: _c_ddf.uint32 | None = ...,
            Reserved1: __cs__.CharArray | None = ...,
            Spare_Type: _c_ddf.uint8 | None = ...,
            Populated_SAEs: _c_ddf.uint16 | None = ...,
            Max_SAE_Supported: _c_ddf.uint16 | None = ...,
            Reserved2: __cs__.CharArray | None = ...,
            Spare_Assignment_Entries: __cs__.Array[Spare_Assignment_Entry] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Physical_Disk_Data(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        PD_GUID: __cs__.CharArray
        PD_Reference: _c_ddf.uint32
        Forced_Ref_Flag: _c_ddf.uint8
        Forced_PD_GUID_Flag: _c_ddf.uint8
        Vendor_Specific_Scratch_Space: __cs__.CharArray
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            PD_GUID: __cs__.CharArray | None = ...,
            PD_Reference: _c_ddf.uint32 | None = ...,
            Forced_Ref_Flag: _c_ddf.uint8 | None = ...,
            Forced_PD_GUID_Flag: _c_ddf.uint8 | None = ...,
            Vendor_Specific_Scratch_Space: __cs__.CharArray | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Mapped_Block_Entry(__cs__.Structure):
        Defective_Block_Start: _c_ddf.uint64
        Spare_Block_Offset: _c_ddf.uint32
        Remapped_Marked_Count: _c_ddf.uint16
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            Defective_Block_Start: _c_ddf.uint64 | None = ...,
            Spare_Block_Offset: _c_ddf.uint32 | None = ...,
            Remapped_Marked_Count: _c_ddf.uint16 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Bad_Block_Management_Log(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Entry_Count: _c_ddf.uint32
        Reserved_Spare_Block_Count: _c_ddf.uint32
        Reserved: __cs__.CharArray
        First_Spare_LBA: _c_ddf.uint64
        class Mapped_Block_Entry(__cs__.Structure):
            Defective_Block_Start: _c_ddf.uint64
            Spare_Block_Offset: _c_ddf.uint32
            Remapped_Marked_Count: _c_ddf.uint16
            Reserved: __cs__.CharArray
            @overload
            def __init__(
                self,
                Defective_Block_Start: _c_ddf.uint64 | None = ...,
                Spare_Block_Offset: _c_ddf.uint32 | None = ...,
                Remapped_Marked_Count: _c_ddf.uint16 | None = ...,
                Reserved: __cs__.CharArray | None = ...,
            ): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        Mapped_Block_Entries: __cs__.Array[Mapped_Block_Entry]
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Entry_Count: _c_ddf.uint32 | None = ...,
            Reserved_Spare_Block_Count: _c_ddf.uint32 | None = ...,
            Reserved: __cs__.CharArray | None = ...,
            First_Spare_LBA: _c_ddf.uint64 | None = ...,
            Mapped_Block_Entries: __cs__.Array[Mapped_Block_Entry] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Vendor_Specific_Log(__cs__.Structure):
        Signature: _c_ddf.uint32
        CRC: _c_ddf.uint32
        Log_Owner: __cs__.CharArray
        Reserved: __cs__.CharArray
        @overload
        def __init__(
            self,
            Signature: _c_ddf.uint32 | None = ...,
            CRC: _c_ddf.uint32 | None = ...,
            Log_Owner: __cs__.CharArray | None = ...,
            Reserved: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_ddf` is an instance of `_c_ddf`, but then we can't use it in type hints
c_ddf: TypeAlias = _c_ddf



from __future__ import annotations

from typing import BinaryIO

from dissect.util.stream import AlignedStream

from dissect.volume.dm.btree import BTree
from dissect.volume.dm.c_dm import SECTOR_SIZE, c_dm
from dissect.volume.exceptions import DMError


class ThinPool:
    """Interact with a device-mapper thin-provision pool."""

    def __init__(self, metadata_fh: BinaryIO, data_fh: BinaryIO):
        self.metadata_fh = metadata_fh
        self.data_fh = data_fh

        self.metadata = Metadata(metadata_fh)

    def open(self, device_id: int, size_hint: int | None = None) -> ThinDevice:
        """Open a thin device on this pool.

        No size information is stored in the pool, so it's recommended to provide a size hint.

        Args:
            device_id: The device ID to open.
            size_hint: Optional size hint for the device to open.
        """
        return ThinDevice(self, device_id, size_hint)


class Metadata:
    """Interact with a device-mapper thin-provision metadata device."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.sb = c_dm.thin_disk_superblock(fh)
        if self.sb.magic != c_dm.THIN_SUPERBLOCK_MAGIC:
            raise DMError("Invalid magic for thin-pool metadata superblock")

        self.data_block_size = self.sb.data_block_size
        self.metadata_block_size = self.sb.metadata_block_size

        self.data_mapping = BTree(self.fh, self.sb.data_mapping_root, self.metadata_block_size)
        self.device_details = BTree(self.fh, self.sb.device_details_root, self.metadata_block_size)


class ThinDevice(AlignedStream):
    """Implements a readable thin device/volume.

    Args:
        pool: The pool the device belongs to.
        device_id: The device ID of the device.
        size_hint: Optional device size hint.
    """

    def __init__(self, pool: ThinPool, device_id: int, size_hint: int | None = None):
        self.pool = pool
        self.device_id = device_id
        self.block_size = self.pool.metadata.data_block_size * SECTOR_SIZE

        details_buf = self.pool.metadata.device_details.lookup(device_id)
        if not details_buf:
            raise DMError(f"Device ID is not known in pool: {device_id}")
        self.details = c_dm.disk_device_details(details_buf)

        super().__init__(size_hint, self.block_size)

    def _read(self, offset: int, length: int) -> bytes:
        data_fh = self.pool.data_fh
        data_mapping = self.pool.metadata.data_mapping

        block = offset // self.block_size

        result = []
        while length > 0:
            block_info = data_mapping.lookup([self.device_id, block])
            if block_info is None:
                remaining = length
                if self.size is None or (remaining := self.size - offset) > 0:
                    return b"\x00" * min(length, remaining)

                break

            block_time = int.from_bytes(block_info, "little")
            data_block, _ = _unpack_block_time(block_time)

            read_size = min(length, self.block_size)
            data_fh.seek(data_block * self.block_size)
            result.append(data_fh.read(read_size))

            length -= read_size
            offset += read_size
            block += 1

        return b"".join(result)


def _unpack_block_time(block_time: int) -> tuple[int, int]:
    """Unpack a block number and timestamp from a combined block time value.

    Block numbers in device-mapper thin-provisioning store the block number in the upper 40 bits and
    a timestamp in the lower 24 bits.
    """
    block = block_time >> 24
    time = block_time & ((1 << 24) - 1)
    return block, time



# http://www.ntfs.com/ldm.htm
#
# This file is still a WIP

from __future__ import annotations

import logging
import os

from dissect.cstruct import cstruct

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_LDM", "CRITICAL"))

ldm_def = """
enum TOCREGION_FLAGS : uint16 {
    NOT_EXIST   = 0x01,
    NEW         = 0x02,
    DELETE      = 0x04,
    DISABLED    = 0x08
};

enum VMDB_STATE : uint16 {
    CONSISTENT  = 0x01,
    CREATION    = 0x02,
    DELETION    = 0x03
};

enum VBLK_STATE : uint16 {
    CONSISTENT  = 0x01,
    DELETION    = 0x02,
    INACTIVE    = 0x03
};

enum VBLK_RECORD_TYPE : uint16 {
    COMPONENT   = 0x32,
    PARTITION   = 0x33,
    DISK3       = 0x34,
    DISK_GROUP3 = 0x35,
    DISK4       = 0x44,
    DISK_GROUP4 = 0x45,
    VOLUME      = 0x51
};

struct PRIVHEAD {
    char        magic[8];
    uint32      checksum;
    uint16      version_major;
    uint16      version_minor;
    uint64      last_update;
    uint64      update_sequence_number;
    uint64      primary_private_header_offset;
    uint64      secondary_private_header_offset;
    char        disk_id[64];
    char        host_id[64];
    char        disk_group_id[64];
    char        disk_group_name[31];
    uint32      sector_size;
    uint32      flags;
    uint16      public_region_slice_number;
    uint16      private_region_slice_number;
    uint64      public_region_offset;
    uint64      public_region_size;
    uint64      primary_toc_offset;
    uint64      secondary_toc_offset;
    uint32      num_configs;
    uint32      num_logs;
    uint64      config_size;
    uint64      log_size;
    char        disk_signature;
    char        disk_set_id[16];
    char        disk_set_id_repeat[16];
};

struct TOCREGION {
    char            name[8];
    TOCREGION_FLAGS flags;
    uint64          offset;
    uint64          size;
    uint16          unk;
    uint16          copy_number;
    char            zeroes[4];
};

struct TOCBLOCK {
    char        magic[8];
    uint32      sequence_number1;
    char        zeroes1[4];
    uint32      sequence_number2;
    char        zeroes2[16];
    TOCREGION   config;
    TOCREGION   log;
};

struct VMDB {
    char        magic[4];
    uint32      last_vblk_sequence;
    uint32      vblk_size;
    uint32      first_vblk;
    VMDB_STATE  update_status;
    uint16      version_major;
    uint16      version_minor;
    char        disk_group_name[31];
    char        disk_group_guid[64];
    uint64      commit_sequence;
    uint64      pending_sequence;
    uint32      num_committed_volume_vblk;
    uint32      num_committed_component_vblk;
    uint32      num_committed_partition_vblk;
    uint32      num_committed_disk_vblk;
    uint32      unused0;
    uint32      unused1;
    uint32      unused2;
    uint32      num_pending_volume_vblk;
    uint32      num_pending_component_vblk;
    uint32      num_pending_partition_vblk;
    uint32      num_pending_disk_vblk;
    uint32      unused3;
    uint32      unused4;
    uint32      unused5;
    uint64      last_access_time;
};

struct VBLK_HEADER {
    char        magic[4];
    uint32      sequence_number;
    uint32      group_number;
    uint16      record_number;
    uint16      num_records;
};

struct VBLK_COMPONENT {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    uint8       volume_state_len;
    char        volume_state[volume_state_len];
    uint8       component_type;
    char        zeroes0[4];
    uint8       num_children_len;
    char        num_children[num_children_len];
    uint64      log_commit_id;
    char        zeroes1[8];
    uint8       parent_id_len;
    char        parent_id[parent_id_len];
    char        zeroes2;
    uint8       strip_size_len;
    char        strip_size[strip_size_len];
    uint8       num_columns_len;
    char        num_columns[num_columns_len];
};

struct VBLK_PARTITION {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    char        zeroes[4];
    uint64      log_commit_id;
    uint64      start;
    uint64      volume_offset;
    uint8       size_len;
    char        size[size_len];
    uint8       parent_id_len;
    char        parent_id[parent_id_len];
    uint8       disk_object_id_len;
    char        disk_object_id[disk_object_id_len];
    uint8       component_part_index_len;
    char        component_part_index[component_part_index_len];
};

struct VBLK_DISK3 {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    uint8       disk_id_len;
    char        disk_id[disk_id_len];
    uint8       alternate_name_len;
    char        alternate_name[alternate_name_len];
    char        zeroes[4];
    uint64      log_commit_id;
};

struct VBLK_DISK4 {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    char        disk_id1[16];
    char        disk_id2[16];
    char        zeroes[3];
    uint16      id;
    uint64      log_commit_id;
};

struct VBLK_DISK_GROUP3 {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    uint8       disk_group_id_len;
    char        disk_group_id[disk_id_len];
    char        zeroes[4];
    uint64      log_commit_id;
    // 0xffffffff
    // 0xffffffff
};

struct VBLK_DISK_GROUP4 {
    uint16      update_status;
    uint16      record_type;
    uint32      data_length;
    uint8       object_id_len;
    char        object_id[objectid_len];
    uint8       name_len;
    char        name[name_len];
    char        disk_group_id[16];
    char        disk_set_id[16];
    char        zeroes[4];
    uint64      log_commit_id;
    // 0xffffffff
    // 0xffffffff
};

struct VBLK_VOLUME {

};

struct KLOG {

};
"""

c_ldm = cstruct(endian=">").load(ldm_def)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_lvm(__cs__.cstruct):
    RAW_LOCN_IGNORED: Literal[1] = ...
    class label_header(__cs__.Structure):
        id: __cs__.CharArray
        sector: _c_lvm.uint64
        crc: _c_lvm.uint32
        offset: _c_lvm.uint32
        type: __cs__.CharArray
        @overload
        def __init__(
            self,
            id: __cs__.CharArray | None = ...,
            sector: _c_lvm.uint64 | None = ...,
            crc: _c_lvm.uint32 | None = ...,
            offset: _c_lvm.uint32 | None = ...,
            type: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class pv_header(__cs__.Structure):
        pv_uuid: __cs__.CharArray
        device_size: _c_lvm.uint64
        @overload
        def __init__(self, pv_uuid: __cs__.CharArray | None = ..., device_size: _c_lvm.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class disk_locn(__cs__.Structure):
        offset: _c_lvm.uint64
        size: _c_lvm.uint64
        @overload
        def __init__(self, offset: _c_lvm.uint64 | None = ..., size: _c_lvm.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class mda_header(__cs__.Structure):
        checksum: _c_lvm.uint32
        magic: __cs__.CharArray
        version: _c_lvm.uint32
        start: _c_lvm.uint64
        size: _c_lvm.uint64
        @overload
        def __init__(
            self,
            checksum: _c_lvm.uint32 | None = ...,
            magic: __cs__.CharArray | None = ...,
            version: _c_lvm.uint32 | None = ...,
            start: _c_lvm.uint64 | None = ...,
            size: _c_lvm.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class raw_locn(__cs__.Structure):
        offset: _c_lvm.uint64
        size: _c_lvm.uint64
        checksum: _c_lvm.uint32
        flags: _c_lvm.uint32
        @overload
        def __init__(
            self,
            offset: _c_lvm.uint64 | None = ...,
            size: _c_lvm.uint64 | None = ...,
            checksum: _c_lvm.uint32 | None = ...,
            flags: _c_lvm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_lvm` is an instance of `_c_lvm`, but then we can't use it in type hints
c_lvm: TypeAlias = _c_lvm

SECTOR_SIZE: int

LABEL_SCAN_SECTORS: int

STATUS_FLAG_ALLOCATABLE: str
STATUS_FLAG_RESIZEABLE: str
STATUS_FLAG_READ: str
STATUS_FLAG_VISIBLE: str
STATUS_FLAG_WRITE: str



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts

from dissect.volume.exceptions import LVM2Error
from dissect.volume.lvm.metadata import VolumeGroup
from dissect.volume.lvm.physical import LVM2Device

if TYPE_CHECKING:
    from datetime import datetime

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_LVM", "CRITICAL"))


class LVM2:
    """Logical Volume Manager."""

    def __init__(self, fh: list[BinaryIO | LVM2Device] | BinaryIO | LVM2Device):
        self.fh = [fh] if not isinstance(fh, list) else fh
        if not self.fh:
            raise ValueError("At least one file-like object is required")

        devices = [LVM2Device(fh) if not isinstance(fh, LVM2Device) else fh for fh in self.fh]
        self.devices = {device.id: device for device in devices}

        self.metadata = devices[0].metadata
        self.contents: str = self.metadata["contents"]
        self.version: int = self.metadata["version"]
        self.description: str | None = self.metadata.get("description")
        self.creation_host: str | None = self.metadata.get("creation_host")
        self.creation_time: datetime | None = None
        if creation_time := self.metadata.get("creation_time"):
            self.creation_time: datetime | None = ts.from_unix(creation_time)

        vg = [VolumeGroup.from_dict(value, name=key) for key, value in self.metadata.items() if isinstance(value, dict)]
        if len(vg) != 1:
            raise LVM2Error(f"Found multiple volume groups, expected only one: {vg}")
        self.volume_group = vg[0]
        self.volume_group.attach(self.devices)

    def __repr__(self) -> str:
        return f"<LVM2 vg={self.vg}>"

    @property
    def vg(self) -> VolumeGroup:
        return self.volume_group



from __future__ import annotations


class Error(Exception):
    pass


class LVM2Error(Error):
    pass


class DiskError(Error):
    pass


class RAIDError(Error):
    pass


class DMError(Error):
    pass


class MDError(Error):
    pass


class DDFError(Error):
    pass



from __future__ import annotations

import logging
import os
import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, TypedDict

from dissect.volume.vinum.c_vinum import c_vinum

if TYPE_CHECKING:
    from datetime import datetime

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_VINUM", "CRITICAL"))


@dataclass
class Volume:
    """The representation of a Vinum Volume.

    A Vinum Volume defines a single RAID set. One or more Vinum Plexes can be
    part of a Volume.
    """

    timestamp: datetime
    name: bytes
    state: VolumeState | None = None


@dataclass
class Plex:
    """The representation of a Vinum Plex.

    A Vinum Plex can be thought of as one of the individual disks in a mirrored
    array. One or more Vinum SDs can be part of a Plex. The Plex defines the
    type of RAID in which these SDs are organized.
    """

    timestamp: datetime
    name: bytes | None = None
    org: PlexOrg | None = None
    stripesize: int | None = None
    volume: bytes | None = None
    state: PlexState | None = None


@dataclass
class SD:
    """The representation of a Vinum SD.

    A Vinum SD contains information about the actual physical disk and points
    to the device of this disk.
    """

    timestamp: datetime
    drive: bytes
    name: bytes | None = None
    # length is the size in bytes of the data section on disk, so without any
    # vinum headers etc.
    length: int | None = None
    # driveoffset is the start of the data section on disk in bytes.
    driveoffset: int | None = None
    plex: bytes | None = None
    # plexoffset is the offset of the data section of this disk within the plex in
    # bytes, e.g.: the first disk always starts at offset 0, if the size of its
    # data section (SD.length) is 1024b then the plexoffset for the second disk
    # will be 1024.
    plexoffset: int | None = None
    state: SDState | None = None


class ParseError(Exception):
    pass


class BytesDefaultEnum(bytes, Enum):
    @classmethod
    def _missing_(cls, value: bytes) -> Any:
        return cls._default


class VolumeState(BytesDefaultEnum):
    DOWN = auto()
    UP = b"up"

    _default = DOWN


class PlexState(BytesDefaultEnum):
    DOWN = auto()
    UP = b"up"
    INITIALIZING = b"initializing"
    DEGRADED = b"degraded"
    GROWABLE = b"growable"

    _default = DOWN


class PlexOrg(BytesDefaultEnum):
    DISORG = auto()
    CONCAT = b"concat"
    STRIPED = b"striped"
    RAID5 = b"raid5"

    _default = DISORG


class SDState(BytesDefaultEnum):
    DOWN = auto()
    UP = b"up"
    INITIALIZING = b"initializing"
    DEGRADED = b"degraded"
    GROWABLE = b"growable"

    _default = DOWN


def _parse_size(size: bytes) -> int:
    # Only the first byte after the numerals (and optional minus sign) should
    # be considered.
    postfix = size.lstrip(b"-0123456789")
    numeral = size[: -len(postfix)] if postfix else size
    unit = postfix[:1]

    try:
        size = int(numeral)
    except ValueError:
        # If there are no numerals (numeral is empty or the minus sign), the
        # size should be parsed as 0.
        size = 0
    else:
        if unit:
            # Invalid unites should be ignored and size is returned as is.
            if unit in (b"b", b"B", b"s", b"S"):
                size = size * 512  # Yes also for b/B
            elif unit in (b"k", b"K"):
                size = size * 1024
            elif unit in (b"m", b"M"):
                size = size * 1024 * 1024
            elif unit in (b"g", b"G"):
                size = size * 1024 * 1024 * 1024

    return size


def _parse_volume_config(config_time: datetime, tokens: list[bytes]) -> Volume | None:
    volume = None
    name = None
    state = None

    tokens = iter(tokens)
    token = next(tokens, None)
    try:
        while token is not None:
            if token == b"state":
                state = VolumeState(next(tokens))
            else:
                name = token
            token = next(tokens, None)
    except StopIteration:
        log.debug("No value for token %r, ignoring volume config", token)
    else:
        if name is None:
            log.debug("No name found for volume, ignoring volume config")
        else:
            volume = Volume(
                timestamp=config_time,
                name=name,
                state=state,
            )

    return volume


def _parse_plex_config(config_time: datetime, tokens: list[bytes]) -> Plex | None:
    plex = None
    name = None
    org = None
    stripesize = None
    volume = None
    state = None

    tokens = iter(tokens)
    token = next(tokens, None)
    try:
        while token is not None:
            if token == b"name":
                name = next(tokens)
            elif token == b"org":
                org = PlexOrg(next(tokens))
                if org == PlexOrg.RAID5 or org == PlexOrg.STRIPED:
                    stripesize = _parse_size(next(tokens))
                    # the kernel parser only checks on == 0, but < 0 also seems unreasonable
                    if stripesize <= 0:
                        raise ParseError(f"Invalid stripesize: {stripesize}")  # noqa: TRY301
            elif token == b"vol" or token == b"volume":
                volume = next(tokens)
            elif token == b"state":
                state = PlexState(next(tokens))
            else:
                raise ParseError(f"Unknown token {token}")  # noqa: TRY301

            token = next(tokens, None)

    except (StopIteration, ParseError) as err:
        if isinstance(err, StopIteration):
            log.debug("No value for token %r, ignoring plex config", token)
        else:
            log.debug("%s, ignoring plex config", err)

    else:
        plex = Plex(
            timestamp=config_time,
            name=name,
            org=org,
            stripesize=stripesize,
            volume=volume,
            state=state,
        )

    return plex


def _parse_sd_config(config_time: datetime, tokens: list[bytes]) -> SD | None:
    sd = None
    name = None
    drive = None
    length = None
    driveoffset = None
    plex = None
    plexoffset = None
    state = None

    tokens = iter(tokens)
    token = next(tokens, None)
    try:
        while token is not None:
            if token == b"name":
                name = next(tokens)
            elif token == b"drive":
                drive = next(tokens)
            elif token == b"len" or token == b"length":
                length = _parse_size(next(tokens))
                if length < 0:
                    length = -1
            elif token == b"driveoffset":
                driveoffset = _parse_size(next(tokens))
                if driveoffset != 0 and driveoffset < c_vinum.GV_DATA_START:
                    raise ParseError(f"Invalid driveoffset: {driveoffset}")  # noqa: TRY301
            elif token == b"plex":
                plex = next(tokens)
            elif token == b"plexoffset":
                plexoffset = _parse_size(next(tokens))
                if plexoffset < 0:
                    raise ParseError(f"Invalid plexoffset: {plexoffset}")  # noqa: TRY301
            elif token == b"state":
                state = SDState(next(tokens))
            else:
                raise ParseError(f"Unknown token {token}")  # noqa: TRY301

            token = next(tokens, None)

    except (StopIteration, ParseError) as err:
        if isinstance(err, StopIteration):
            log.debug("No value for token %r, ignoring sd config", token)
        else:
            log.debug("%s, ignoring sd config", err)

    else:
        if drive is None:
            log.debug("No drive found for sd, ignoring sd config")
        else:
            sd = SD(
                timestamp=config_time,
                name=name,
                drive=drive,
                length=length,
                driveoffset=driveoffset,
                plex=plex,
                plexoffset=plexoffset,
                state=state,
            )

    return sd


def get_char(line: bytes, idx: int) -> bytes:
    """Return a single byte bytestring at index ``idx`` in ``line``.

    If the index is outside of the bounaries of ``line``, an empty bytestring
    will be returned.
    """
    char = b""
    if idx >= 0 and idx < len(line):
        char = line[idx : idx + 1]  # this makes sure we get a single byte bytestring
    return char


class TokenizeError(Exception):
    pass


def tokenize(line: bytes) -> iter[bytes]:
    """Yield individual tokens from a vinum config line.

    This token parser is constructed to be equivalent to the token parser used in the
    FreeBSD kernel code. There are a few caveats though:

    - it expects lines to be pre-splitted on newline and null-byte characters
    - it does not attempt to parse quoted tokens, as the code in the kernel parser is
      buggy and will always lead to an error condition (it will mimick the error condition
      though).
    """
    whitespace = {b" ", b"\t"}
    quotes = {b'"', b"'"}
    comment = {b"#"}
    eol = {b""}
    end_of_list = eol.union(comment)
    end_of_token = whitespace.union(eol)

    token = b""
    idx = 0
    while True:
        char = get_char(line, idx)

        while char in whitespace:
            # Remove leading whitespace up to the next token or end_of_list condition
            idx += 1
            char = get_char(line, idx)

        if char in end_of_list:
            # We are at the end of the token list (a comment or end of line).
            break

        if char in quotes:
            # Encountering a quoted token will always lead to an error
            # condition in the (Free)BSD vinum kernel code. This is a bug in
            # that code, which we mimick here.
            raise TokenizeError(f"Found quoted token at index {idx}")

        while char not in end_of_token:
            # Add characters to the token until we encounter a stop condition.
            # Note that comment and quote characters are allowed in a token as
            # long as they are not preceded by whitespace.
            token += char
            idx += 1
            char = get_char(line, idx)

        if token:
            yield token
            token = b""

        idx += 1


class VinumConfigs(TypedDict):
    """Vinum configurations."""

    volumes: list[Volume]
    plexes: list[Plex]
    sds: list[SD]


RE_CONFIG_EOL = re.compile(b"[\x00\n]")


TOKEN_CONFIG_MAP = {
    b"volume": "volumes",
    b"plex": "plexes",
    b"sd": "sds",
}


def parse_vinum_config(config_time: datetime, config: bytes) -> VinumConfigs:
    """Parse the on-disk vinum configuration.

    Parsing forgiveness and strictness is implemented in the same way as in the vinum kernel code:

    Lines with an unknown configuration ``type`` (not ``volume``, ``plex`` or ``sd``), are ignored.

    Lines that fail to parse due to:
      - no name present
      - no value present for a token
      - unknown token name
      - a tokenization error

    will fail that line and the subsequent lines (rest of the config) to not being parsed.
    """
    config_data: VinumConfigs = {
        "volumes": [],
        "plexes": [],
        "sds": [],
    }

    for line in RE_CONFIG_EOL.split(config):
        try:
            tokens = tokenize(line)
            token = next(tokens, None)
            if token is None:
                # We encountered a line without tokens (empty, just whitespace or # comments)
                continue
            if token == b"volume":
                parsed_config = _parse_volume_config(config_time, tokens)
            elif token == b"plex":
                parsed_config = _parse_plex_config(config_time, tokens)
            elif token == b"sd":
                parsed_config = _parse_sd_config(config_time, tokens)
            else:
                parsed_config = None
                log.debug("Unknown config type in line: %r, ignoring config line", line)

            if parsed_config:
                config_type = TOKEN_CONFIG_MAP[token]
                config_data[config_type].append(parsed_config)
            else:
                log.debug("Invalid config line %r", line)
                log.debug("Ignoring this line and the rest of the config data")
                break
        except TokenizeError as err:
            log.debug("Invalid config line %r: %s", line, err)
            log.debug("Ignoring this line and the rest of the config data")
            break

    return config_data



# Reference: https://www.snia.org/sites/default/files/SNIA_DDF_Technical_Position_v2.0.pdf
from __future__ import annotations

from dissect.cstruct import cstruct

ddf_def = """
#define DDF_HEADER_SIGNATURE            0xDE11DE11
#define DDF_HEADER_CLOSED               0x00
#define DDF_HEADER_OPENED_MASK          0x0F
#define DDF_HEADER_OPEN_ANCHOR          0xFF
#define DDF_HEADER_ANCHOR               0x00
#define DDF_HEADER_PRIMARY              0x01
#define DDF_HEADER_SECONDARY            0x02

struct DDF_Header {
    uint32  Signature;
    uint32  CRC;
    char    DDF_Header_GUID[24];
    char    DDF_rev[8];
    uint32  Sequence_Number;
    uint32  TimeStamp;
    uint8   Open_Flag;
    uint8   Foreign_Flag;
    uint8   Disk_Grouping;
    char    Reserved1[13];
    uint8   Header_Ext[32];
    uint64  Primary_Header_LBA;
    uint64  Secondary_Header_LBA;
    uint8   Header_Type;
    char    Reserved2[3];
    uint32  Workspace_Length;
    uint64  Workspace_LBA;
    uint16  Max_PD_Entries;
    uint16  Max_VD_Entries;
    uint16  Max_Partitions;
    uint16  Configuration_Record_Length;
    uint16  Max_Primary_Element_Entries;
    uint32  Max_Mapped_Block_Entries;                       /* DDF 2.0 */
    char    Reserved3[50];
    uint32  Controller_Data_Section;
    uint32  Controller_Data_Section_Length;
    uint32  Physical_Disk_Records_Section;
    uint32  Physical_Disk_Records_Section_Length;
    uint32  Virtual_Disk_Records_Section;
    uint32  Virtual_Disk_Records_Section_Length;
    uint32  Configuration_Records_Section;
    uint32  Configuration_Records_Section_Length;
    uint32  Physical_Disk_Data_Section;
    uint32  Physical_Disk_Data_Section_Length;
    uint32  BBM_Log_Section;
    uint32  BBM_Log_Section_Length;
    uint32  Diagnostic_Space_Section;
    uint32  Diagnostic_Space_Section_Length;
    uint32  Vendor_Specific_Logs_Section;
    uint32  Vendor_Specific_Logs_Section_Length;
    char    Reserved4[256];
};

#define DDF_CONTROLLER_DATA_SIGNATURE   0xAD111111

struct Controller_Data {
    uint32  Signature;
    uint32  CRC;
    char    Controller_GUID[24];
    char    Controller_Type[8];
    char    Product_ID[16];
    char    Reserved[8];
    char    Vendor_Unique_Controller_Data[448];
};

#define DDF_PDR_SIGNATURE               0x22222222

#define DDF_PDE_GUID_FORCE              (1 << 0)
#define DDF_PDE_PARTICIPATING           (1 << 1)
#define DDF_PDE_GLOBAL_SPARE            (1 << 2)
#define DDF_PDE_CONFIG_SPARE            (1 << 3)
#define DDF_PDE_FOREIGN                 (1 << 4)
#define DDF_PDE_LEGACY                  (1 << 5)
#define DDF_PDE_TYPE_MASK               (0x0f << 12)
#define DDF_PDE_UNKNOWN                 (0x00 << 12)
#define DDF_PDE_SCSI                    (0x01 << 12)
#define DDF_PDE_SAS                     (0x02 << 12)
#define DDF_PDE_SATA                    (0x03 << 12)
#define DDF_PDE_FC                      (0x04 << 12)

#define DDF_PDE_ONLINE                  (1 << 0)
#define DDF_PDE_FAILED                  (1 << 1)
#define DDF_PDE_REBUILD                 (1 << 2)
#define DDF_PDE_TRANSITION              (1 << 3)
#define DDF_PDE_PFA                     (1 << 4)
#define DDF_PDE_UNRECOVERED             (1 << 5)
#define DDF_PDE_MISSING                 (1 << 6)

struct Physical_Disk_Entry {
    char    PD_GUID[24];
    uint32  PD_Reference;
    uint16  PD_Type;
    uint16  PD_State;
    uint64  Configured_Size;
    char    Path_Information[18];
    uint16  Block_Size;                                     /* DDF 2.0 */
    char    Reserved[4];
};

struct Physical_Disk_Records {
    uint32  Signature;
    uint32  CRC;
    uint16  Populated_PDEs;
    uint16  Max_PDE_Supported;
    char    Reserved[52];
    // Variable data
    // Physical_Disk_Entry Physical_Disk_Entries[Populated_PDEs];
};

#define DDF_VD_RECORD_SIGNATURE         0xDDDDDDDD

#define DDF_VDE_SHARED                  (1 << 0)
#define DDF_VDE_ENFORCE_GROUP           (1 << 1)
#define DDF_VDE_UNICODE_NAME            (1 << 2)
#define DDF_VDE_OWNER_ID_VALID          (1 << 3)

#define DDF_VDE_OPTIMAL                 0x00
#define DDF_VDE_DEGRADED                0x01
#define DDF_VDE_DELETED                 0x02
#define DDF_VDE_MISSING                 0x03
#define DDF_VDE_FAILED                  0x04
#define DDF_VDE_PARTIAL                 0x05
#define DDF_VDE_OFFLINE                 0x06
#define DDF_VDE_STATE_MASK              0x07
#define DDF_VDE_MORPH                   (1 << 3)
#define DDF_VDE_DIRTY                   (1 << 4)

#define DDF_VDE_UNINTIALIZED            0x00
#define DDF_VDE_INIT_QUICK              0x01
#define DDF_VDE_INIT_FULL               0x02
#define DDF_VDE_INIT_MASK               0x03
#define DDF_VDE_UACCESS_RW              0x00
#define DDF_VDE_UACCESS_RO              0x80
#define DDF_VDE_UACCESS_BLOCKED         0xc0
#define DDF_VDE_UACCESS_MASK            0xc0

struct Virtual_Disk_Entry {
    char    VD_GUID[24];
    uint16  VD_Number;
    char    Reserved1[2];
    uint32  VD_Type;
    uint8   VD_State;
    uint8   Init_State;
    uint8   Partially_Optimal_Drive_Failures_Remaining;     /* DDF 2.0 */
    char    Reserved2[13];
    char    VD_Name[16];
};

struct Virtual_Disk_Records {
    uint32  Signature;
    uint32  CRC;
    uint16  Populated_VDEs;
    uint16  Max_VDE_Supported;
    char    Reserved[52];
    // Variable data
    // Virtual_Disk_Entry  Virtual_Disk_Entries[Populated_VDEs];
};

#define DDF_VDCR_SIGNATURE              0xEEEEEEEE

/* Primary Raid Level (PRL) */
#define DDF_VDCR_RAID0                  0x00
#define DDF_VDCR_RAID1                  0x01
#define DDF_VDCR_RAID3                  0x03
#define DDF_VDCR_RAID4                  0x04
#define DDF_VDCR_RAID5                  0x05
#define DDF_VDCR_RAID6                  0x06
#define DDF_VDCR_RAID1E                 0x11
#define DDF_VDCR_SINGLE                 0x0F
#define DDF_VDCR_CONCAT                 0x1F
#define DDF_VDCR_RAID5E                 0x15
#define DDF_VDCR_RAID5EE                0x25

/* Raid Level Qualifier (RLQ) */
#define DDF_VDCR_RAID0_SIMPLE           0x00
#define DDF_VDCR_RAID1_SIMPLE           0x00
#define DDF_VDCR_RAID1_MULTI            0x01
#define DDF_VDCR_RAID3_0                0x00
#define DDF_VDCR_RAID3_N                0x01
#define DDF_VDCR_RAID4_0                0x00
#define DDF_VDCR_RAID4_N                0x01
#define DDF_VDCR_RAID5_0_RESTART        0x00
#define DDF_VDCR_RAID6_0_RESTART        0x01
#define DDF_VDCR_RAID5_N_RESTART        0x02
#define DDF_VDCR_RAID5_N_CONTINUE       0x03

#define DDF_VDCR_RAID1E_ADJACENT        0x00
#define DDF_VDCR_RAID1E_OFFSET          0x01

/* Secondary RAID Level (SRL) */
#define DDF_VDCR_2STRIPED               0x00
#define DDF_VDCR_2MIRRORED              0x01
#define DDF_VDCR_2CONCAT                0x02
#define DDF_VDCR_2SPANNED               0x03

#define DDF_VDCR_CACHE_WB               (1 << 0)
#define DDF_VDCR_CACHE_WB_ADAPTIVE      (1 << 1)
#define DDF_VDCR_CACHE_RA               (1 << 2)
#define DDF_VDCR_CACHE_RA_ADAPTIVE      (1 << 3)
#define DDF_VDCR_CACHE_WCACHE_NOBATTERY (1 << 4)
#define DDF_VDCR_CACHE_WCACHE_ALLOW     (1 << 5)
#define DDF_VDCR_CACHE_RCACHE_ALLOW     (1 << 6)
#define DDF_VDCR_CACHE_VENDOR           (1 << 7)

struct VD_Configuration_Record {
    uint32  Signature;
    uint32  CRC;
    char    VD_GUID[24];
    uint32  Timestamp;
    uint32  Sequence_Number;
    char    Reserved1[24];
    uint16  Primary_Element_Count;
    uint8   Strip_Size;
    uint8   Primary_RAID_Level;
    uint8   RAID_Level_Qualifier;
    uint8   Secondary_Element_Count;
    uint8   Secondary_Element_Seq;
    uint8   Secondary_RAID_Level;
    uint64  Block_Count;
    uint64  VD_Size;
    uint16  Block_Size;                                     /* DDF 2.0 */
    uint8   Rotate_Parity_Count;                            /* DDF 2.0 */
    char    Reserved2[5];
    char    Associated_Spares[32];
    uint64  Cache_Policies_And_Parameters;
    uint8   BG_Rate;
    char    Reserved3[3];
    uint8   MDF_Parity_Disks;                               /* DDF 2.0 */
    uint16  MDF_Parity_Generator_Polynomial;                /* DDF 2.0 */
    char    Reserved4[1];
    uint8   MDF_Constant_Generation_Method;                 /* DDF 2.0 */
    char    Reserved5[47];
    char    Reserved6[192];
    char    V0[32];
    char    V1[32];
    char    V2[16];
    char    V3[16];
    char    Vendor_Specific_Scratch_Space[32];
    // Variable data
    // uint32  Physical_Disk_Sequence[0];
    // uint64  Starting_Block[0];
};

#define DDF_VUCR_SIGNATURE              0x88888888

struct VU_Configuration_Record {
    uint32  Signature;
    uint32  CRC;
    char    VD_GUID[24];
};

#define DDF_SA_SIGNATURE                0x55555555

struct Spare_Assignment_Entry {
    char    VD_GUID[24];
    uint16  Secondary_Element;
    char    Reserved[6];
};

#define DDF_SAR_TYPE_DEDICATED          (1 << 0)
#define DDF_SAR_TYPE_REVERTIBLE         (1 << 1)
#define DDF_SAR_TYPE_ACTIVE             (1 << 2)
#define DDF_SAR_TYPE_ENCL_AFFINITY      (1 << 3)

struct Spare_Assignment_Record {
    uint32  Signature;
    uint32  CRC;
    uint32  Timestamp;
    char    Reserved1[7];
    uint8   Spare_Type;
    uint16  Populated_SAEs;
    uint16  Max_SAE_Supported;
    char    Reserved2[8];
    Spare_Assignment_Entry  Spare_Assignment_Entries[Populated_SAEs];
};

#define DDF_PDD_SIGNATURE               0x33333333

#define DDF_PDD_FORCED_REF              0x01
#define DDF_PDD_FORCED_GUID             0x01

struct Physical_Disk_Data {
    uint32  Signature;
    uint32  CRC;
    char    PD_GUID[24];
    uint32  PD_Reference;
    uint8   Forced_Ref_Flag;
    uint8   Forced_PD_GUID_Flag;
    char    Vendor_Specific_Scratch_Space[32];
    char    Reserved[442];
};

#define DDF_BBML_SIGNATURE              0xABADB10C

struct Mapped_Block_Entry {
    uint64  Defective_Block_Start;
    uint32  Spare_Block_Offset;
    uint16  Remapped_Marked_Count;
    char    Reserved[2];
};

struct Bad_Block_Management_Log {
    uint32  Signature;
    uint32  CRC;
    uint32  Entry_Count;
    uint32  Reserved_Spare_Block_Count;
    char    Reserved[8];
    uint64  First_Spare_LBA;
    Mapped_Block_Entry  Mapped_Block_Entries[Entry_Count];
};

#define DDF_VENDOR_LOG_SIGNATURE        0x01DBEEF0

struct Vendor_Specific_Log {
    uint32  Signature;
    uint32  CRC;
    char    Log_Owner[8];
    char    Reserved[16];
};
"""
c_ddf = cstruct(endian=">").load(ddf_def)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.volume.disk.partition import Partition
from dissect.volume.exceptions import DiskError

if TYPE_CHECKING:
    from collections.abc import Iterator

mbr_def = """
typedef struct part_s {
    uint8   bootable;       // +0: 0x80/0x00 - bootable/not bootable
    uint8   start_head;     // +1: head (start)
    uint16  start_cyl_sec;  // +2: cyl+sect (start)
    uint8   type;           // +4: type
    uint8   end_head;       // +5: head (end)
    uint16  end_cyl_sec;    // +6: cyl+sec (end)
    uint32  sector_ofs;     // +8: offset in sectors
    uint32  sector_size;    // +12: size in sectors
} part;

typedef struct mbr_s {
    char    bootcode[0x1b8];
    uint32  vol_no;
    uint16  pad1;
    part    part[4];
    uint16  bootsig;
} mbr;
"""

c_mbr = cstruct().load(mbr_def)


class MBR:
    """Master Boot Record."""

    def __init__(self, fh: BinaryIO, sector_size: int = 512):
        self.fh = fh
        self.sector_size = sector_size
        self.offset = fh.tell()
        self.mbr = c_mbr.mbr_s(fh)

        if self.mbr.bootsig != 0xAA55:
            raise DiskError(f"Invalid MBR boot signature. Expected 0xaa55, got 0x{self.mbr.bootsig:x}.")

        # This sucks but don't have a better way atm
        sig = self.mbr.bootcode[3:11]
        if any(v in sig for v in [b"MSDOS", b"MSWIN", b"NTFS", b"FAT", b"EXFAT", b"-FVE-FS-", b"SYSLINUX"]):
            raise DiskError("Sector is a filesystem VBR, not an MBR")

        if self.mbr.bootcode[18:38] == b"Hit Esc for .altboot" or self.mbr.bootcode[168:174] == b"\r\nQNX ":
            raise DiskError("Sector is a QNX Boot Sector, not an MBR")

        self.partitions: list[Partition] = list(self._partitions(self.mbr, self.offset))

    def _partitions(
        self, mbr: c_mbr.mbr_s, offset: int, num_start: int = 0, ebr_offset: int | None = None
    ) -> Iterator[Partition]:
        for num, partition in enumerate(mbr.part):
            if partition.type == 0x00:
                continue

            part_offset = offset + partition.sector_ofs * self.sector_size

            if partition.type in (0x05, 0x0F, 0x85):  # Extended
                if not ebr_offset:
                    ebr_offset = part_offset
                else:
                    # Chained extended MBRs are relative to the first one
                    part_offset = ebr_offset + partition.sector_ofs * self.sector_size

                self.fh.seek(part_offset)
                e_mbr = c_mbr.mbr_s(self.fh)
                yield from self._partitions(e_mbr, part_offset, num_start + num, ebr_offset=ebr_offset)

                continue

            yield Partition(
                disk=self,
                number=num_start + num + 1,  # partitions are 1-indexed
                offset=part_offset,
                size=partition.sector_size * self.sector_size,
                type=partition.type,
                flags=partition.bootable,
                raw=partition,
            )



from __future__ import annotations

import ast
import re
from bisect import bisect_right
from functools import cached_property
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.util.stream import RunlistStream

from dissect.volume.exceptions import LVM2Error
from dissect.volume.lvm.c_lvm2 import LABEL_SCAN_SECTORS, SECTOR_SIZE, c_lvm

if TYPE_CHECKING:
    from collections.abc import Iterator


class LVM2Device:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        for i in range(LABEL_SCAN_SECTORS):
            fh.seek(i * SECTOR_SIZE)
            lbl = c_lvm.label_header(fh)
            if lbl.id == b"LABELONE":
                self.label = lbl
                self.label_offset = i * SECTOR_SIZE
                break
        else:
            raise LVM2Error("Can't find physical volume label header")

        fh.seek(self.label_offset + self.label.offset)

        self.header = c_lvm.pv_header(fh)
        self.id = self.header.pv_uuid.decode()
        self.size = self.header.device_size

        self._data_area_descriptors = _read_descriptors(fh, c_lvm.disk_locn)

        if len(self._data_area_descriptors) != 1:
            # According to the LVM2 code, there should only be 1 data area descriptor for each Physical Volume
            raise RuntimeError(f"There should be 1 data area descriptor, found {len(self._data_area_descriptors)}")

        self._metadata_area_descriptors = _read_descriptors(fh, c_lvm.disk_locn)

        self._metadata_areas = []
        for desc in self._metadata_area_descriptors:
            self.fh.seek(desc.offset)
            header = c_lvm.mda_header(self.fh)
            raw_location_descriptors = _read_descriptors(self.fh, c_lvm.raw_locn)
            self._metadata_areas.append((header, raw_location_descriptors))

        self._data_area_offsets = []

        offset = 0
        for area in self._data_area_descriptors:
            if offset != 0:
                self._data_area_offsets.append(offset)
            offset += area.size

    def __repr__(self) -> str:
        return f"<LVMDevice id={self.id}, size={self.size:#x}>"

    @cached_property
    def metadata(self) -> dict | None:
        if not self._metadata_areas:
            return None

        # Only support the first metadata area for now
        # Unsure when a second area is used
        mda_header, raw_locn = self._metadata_areas[0]

        self.fh.seek(mda_header.start + raw_locn[0].offset)
        return parse_metadata(self.fh.read(raw_locn[0].size - 1).decode())

    def read_sectors(self, sector: int, count: int) -> bytes:
        r = []
        area_idx = bisect_right(self._data_area_offsets, sector * SECTOR_SIZE)

        while count > 0:
            area = self._data_area_descriptors[area_idx]

            area_size = area.size or self.size
            area_remaining_sectors = (area_size // SECTOR_SIZE) - (sector - area.offset // SECTOR_SIZE)
            read_sectors = min(area_remaining_sectors, count)

            self.fh.seek(area.offset + sector * SECTOR_SIZE)
            r.append(self.fh.read(read_sectors * SECTOR_SIZE))
            sector += read_sectors
            count -= read_sectors

            area_idx += 1

        return b"".join(r)

    def open(self) -> BinaryIO:
        # Only accounts for one data area descriptor
        # When multiple descriptors get allowed in the future, this code should be adjusted to reflect that
        area = self._data_area_descriptors[0]
        runlist = [(area.offset // SECTOR_SIZE, self.size // SECTOR_SIZE)]
        return RunlistStream(self.fh, runlist, self.size, SECTOR_SIZE)


def _read_descriptors(
    fh: BinaryIO, ctype: type[c_lvm.disk_locn | c_lvm.raw_locn]
) -> list[c_lvm.disk_locn | c_lvm.raw_locn]:
    # Read an unspecified number of descriptors from the file handle, stopping on all zeros
    return ctype[None](fh)


def parse_metadata(string: str) -> dict:
    root = {}
    current = root
    parents = {}

    s = re.sub(r"(#[^\"]+?)$", "", string, flags=re.MULTILINE)

    it = iter(s.split("\n"))
    for line in it:
        line = line.strip()
        if not line or line[0] == "#":
            continue

        if line[-1] == "{":
            name = line[:-1].strip()

            child = {}
            parent = current
            parents[id(child)] = parent
            parent[name] = child
            current = child
            continue

        if line[-1] == "}":
            current = parents[id(current)]
            continue

        k, v = _parse_key_value(line, it)
        current[k] = v

    root["_raw"] = string
    return root


def _parse_key_value(s: str, it: Iterator[str]) -> tuple[str, Any]:
    key, value = s.strip().split("=", 1)
    key = key.strip()
    value = value.strip()

    if value[0] == "[" and value[-1] != "]":
        lines = [value]
        for line in it:
            lines.append(line)
            if line[-1] == "]":
                break
        value = "".join(lines)

    value = ast.literal_eval(value)

    return key, value



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_vinum(__cs__.cstruct):
    GV_HDR_OFFSET: Literal[4096] = ...
    GV_HDR_LEN: Literal[512] = ...
    GV_CFG_OFFSET: Literal[4608] = ...
    GV_CFG_LEN: Literal[65536] = ...
    GV_DATA_START: Literal[135680] = ...
    GV_MAXDRIVENAME: Literal[32] = ...
    GV_HOSTNAME_LEN: Literal[32] = ...
    GV_OLD_MAGIC: Literal[5282194967921315584] = ...
    GV_OLD_NOMAGIC: Literal[5642764413087665920] = ...
    GV_MAGIC: Literal[6217586888835674368] = ...
    GV_NOMAGIC: Literal[6217586888835673344] = ...
    class timeval(__cs__.Structure):
        sec: _c_vinum.uint64
        usec: _c_vinum.uint64
        @overload
        def __init__(self, sec: _c_vinum.uint64 | None = ..., usec: _c_vinum.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    off_t: TypeAlias = _c_vinum.uint64
    class gv_label(__cs__.Structure):
        sysname: __cs__.CharArray
        name: __cs__.CharArray
        date_of_birth: _c_vinum.timeval
        last_update: _c_vinum.timeval
        drive_size: _c_vinum.uint64
        @overload
        def __init__(
            self,
            sysname: __cs__.CharArray | None = ...,
            name: __cs__.CharArray | None = ...,
            date_of_birth: _c_vinum.timeval | None = ...,
            last_update: _c_vinum.timeval | None = ...,
            drive_size: _c_vinum.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class gv_hdr(__cs__.Structure):
        magic: _c_vinum.uint64
        config_length: _c_vinum.uint64
        label: _c_vinum.gv_label
        @overload
        def __init__(
            self,
            magic: _c_vinum.uint64 | None = ...,
            config_length: _c_vinum.uint64 | None = ...,
            label: _c_vinum.gv_label | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    header: TypeAlias = gv_hdr

# Technically `c_vinum` is an instance of `_c_vinum`, but then we can't use it in type hints
c_vinum: TypeAlias = _c_vinum

SECTOR_SIZE: int

MAGIC_ACTIVE: set[int]
MAGIC_INACTIVE: set[int]
